/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Properties;
import java.util.TimeZone;

import org.w3c.dom.Document;

import de.todesbaum.freemule.util.DateFormatter;
import de.todesbaum.freemule.util.Hex;
import de.todesbaum.freemule.util.SimpleXML;
import de.todesbaum.freemule.util.XML;
import de.todesbaum.util.freenet.fcp.FcpException;
import de.todesbaum.util.freenet.fcp.FcpKeyCollisionException;
import de.todesbaum.util.freenet.fcp.FcpRuntimeException;


/**
 * The Announcement Service creates announcement information and inserts it into
 * the net, inserting a fresh copy with the latest information every hour.
 * 
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:AnnouncementService.java 220 2006-03-06 14:46:23Z bombe $
 */
public class AnnouncementService extends AbstractEngineService {

	public static final String NAME = "Announcement Service";

	/**
	 * A {@link List} of {@link AnnouncementListener}s.
	 */
	private List<AnnouncementListener> announcementListeners;

	/**
	 * The local client.
	 */
	private Client client;

	/**
	 * The GMT timezone, used for inserting at the correct path.
	 */
	private TimeZone gmt;

	/**
	 * Whether announcement information was already inserted this hour.
	 */
	private boolean insertedThisHour;

	/**
	 * The last value of the current hour, used to detect a change in the hour
	 * value of the current time.
	 */
	private int lastCurrentHour;

	/**
	 * Whether to insert new announcement information now.
	 */
	private boolean insertNow;

	/**
	 * The most current filelist index to insert.
	 */
	private int lastFilelistIndex;

	/**
	 * {@inheritDoc}
	 */
	public String getName() {
		return NAME;
	}

	/**
	 * Adds an {@link AnnouncementListener} to the list of registered listeners.
	 * 
	 * @param announcementListener
	 *            The listener to add
	 */
	public void addAnnouncementListener(AnnouncementListener announcementListener) {
		if (!announcementListeners.contains(announcementListener)) {
			announcementListeners.add(announcementListener);
		}
	}

	/**
	 * Removes an {@link AnnouncementListener} from the list of registered
	 * listeners.
	 * 
	 * @param announcementListener
	 *            The listener to remove
	 */
	public void removeAnnouncementListener(AnnouncementListener announcementListener) {
		announcementListeners.remove(announcementListener);
	}

	/**
	 * Notifies listeners that an announcement insert has been started.
	 * 
	 * @param key
	 *            The key of the announcement
	 */
	private void fireAnnouncementInsertStarted(String key) {
		for (AnnouncementListener announcementListener: announcementListeners) {
			announcementListener.announcementInsertStarted(key);
		}
	}

	/**
	 * Notifies listeners that the announcement could not be inserted at the
	 * specified key.
	 * 
	 * @see AnnouncementListener#announcementInsertFailed(String, Exception)
	 * @param key
	 *            The key the announcement was about to be inserted to
	 * @param exception
	 *            The exception that prevented insertion
	 */
	private void fireAnnouncementInsertFailed(String key, Exception exception) {
		for (AnnouncementListener announcementListener: announcementListeners) {
			announcementListener.announcementInsertFailed(key, exception);
		}
	}

	/**
	 * Notifies listeners that the announcement was successfully inserted to the
	 * specified key.
	 * 
	 * @see AnnouncementListener#announcementInsertComplete(String)
	 * @param key
	 *            The key the announcement was inserted to
	 */
	private void fireAnnouncementInsertComplete(String key) {
		for (AnnouncementListener announcementListener: announcementListeners) {
			announcementListener.announcementInsertComplete(key);
		}
	}

	/**
	 * Sets the local client.
	 * 
	 * @param client
	 *            The local client
	 */
	public void setClient(Client client) {
		this.client = client;
	}

	/**
	 * Trigger an insert with the new filelist index.
	 * 
	 * @param filelistIndex
	 *            The new filelist index to insert
	 */
	public void insertNow(int filelistIndex) {
		synchronized (syncObject) {
			lastFilelistIndex = filelistIndex;
			insertNow = true;
			syncObject.notify();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void init() {
		super.init();
		announcementListeners = new ArrayList<AnnouncementListener>();
		client = null;
		gmt = TimeZone.getTimeZone("GMT-00:00");
		insertedThisHour = false;
		lastCurrentHour = -1;
		insertNow = false;
	}

	/**
	 * {@inheritDoc}
	 */
	public void destroy() {
		announcementListeners.clear();
		announcementListeners = null;
		client = null;
		gmt = null;
		super.destroy();
	}

	private int getCurrentHour() {
		Calendar calendar = Calendar.getInstance(gmt);
		return calendar.get(Calendar.HOUR_OF_DAY);
	}

	/**
	 * The real work-horse. Inserts new announcement information as soon as a
	 * change of hours was detected, or if {@link #insertNow} has been set by
	 * {@link #insertNow(int)}.
	 */
	protected void work() {
		Calendar calendar = Calendar.getInstance(gmt);
		int currentHour = getCurrentHour();
		if (currentHour != lastCurrentHour) {
			lastCurrentHour = currentHour;
			insertedThisHour = false;
		}

		synchronized (syncObject) {
			insertedThisHour &= !insertNow;
		}

		if (!insertedThisHour) {
			Document announcement;
			synchronized (syncObject) {
				announcement = createAnnouncement(lastFilelistIndex);
			}
			byte[] announcementContent = XML.transformToByteArray(announcement);
			Data announcementData = new Data(new byte[0], announcementContent);
			String keyPrefix = "KSK@" + networkName + "/announcement/" + DateFormatter.getKeyPortion(calendar) + "/";
			int index = 0;
			String key = null;
			while (!stopping && !insertedThisHour && (getCurrentHour() == lastCurrentHour)) {
				try {
					key = keyPrefix + Hex.toHex(index);
					setStatus("Inserting Announcement at " + key + "...");
					fireAnnouncementInsertStarted(key);
					engine.putKey(key, hopsToLive, announcementData);
					fireAnnouncementInsertComplete(key);
					insertedThisHour = true;
					synchronized (syncObject) {
						insertNow = false;
					}
				} catch (FcpKeyCollisionException fkce1) {
					index++;
				} catch (FcpRuntimeException fre1) {
					fireAnnouncementInsertFailed(key, fre1);
				} catch (FcpException fe1) {
					fireServiceAborted(fe1);
					stopping = true;
					break;
				} catch (IOException ioe1) {
					fireAnnouncementInsertFailed(key, ioe1);
				} catch (EngineStoppedException ese1) {
					fireServiceAborted(ese1);
					stopping = true;
					break;
				}
			}
		}
	}

	/**
	 * Creates announcement information.
	 * 
	 * @param index
	 *            The filelist index
	 * @return A {@link Properties} object containing announcement information
	 */
	private Document createAnnouncement(int index) {
		SimpleXML announcementXML = new SimpleXML("announcement");
		announcementXML.append("client-id", client.getClientId());
		announcementXML.append("client-name", client.getName());
		announcementXML.append("client-version", client.getVersion());
		announcementXML.append("client-address", client.getAddress());
		announcementXML.append("client-filelist-index", Hex.toHex(index));
		announcementXML.append("client-timestamp", Hex.toHex(client.getTimestamp()));
		return announcementXML.getDocument();
	}

}
