/*
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.StringTokenizer;

import org.w3c.dom.Document;

import de.todesbaum.freemule.util.Hex;
import de.todesbaum.freemule.util.SimpleXML;
import de.todesbaum.util.freenet.fcp.FcpClient;
import de.todesbaum.util.freenet.fcp.FcpClientGet;
import de.todesbaum.util.freenet.fcp.FcpClientHello;
import de.todesbaum.util.freenet.fcp.FcpClientPut;
import de.todesbaum.util.freenet.fcp.FcpCommand;
import de.todesbaum.util.freenet.fcp.FcpException;
import de.todesbaum.util.freenet.fcp.FcpGenerateCHK;
import de.todesbaum.util.freenet.fcp.FcpGenerateSVKPair;
import de.todesbaum.util.freenet.fcp.FcpKeyCollisionException;
import de.todesbaum.util.freenet.fcp.FcpNode;
import de.todesbaum.util.freenet.fcp.FcpRedirect;

/**
 * The EngineService is the interface that freeMule uses to communicate with the Freenet Node.
 * 
 * @author <a href="mailto:droden@gmail.com">David Roden </a>
 * @version $Id:EngineService.java 220 2006-03-06 14:46:23Z bombe $
 */
public class EngineService extends AbstractService {

	public static final String NAME = "Engine Service";

	/**
	 * The cache used by the engine.
	 * <p>
	 * <strong>Note:</strong> The cache is disabled in the current implementation!
	 */
	private Cache cache;

	/**
	 * The {@link FcpClient} used to communicate with the node.
	 */
	private FcpClient fcpClient;

	/**
	 * Whether the engine is currently running.
	 */
	private boolean running = false;

	/**
	 * {@inheritDoc}
	 */
	public String getName() {
		return NAME;
	}

	/**
	 * Sets the {@link FcpNode} to use.
	 * 
	 * @param fcpNode
	 *            The fcpNode to use
	 */
	public void setFcpNode(FcpNode fcpNode) {
		fcpClient = new FcpClient(fcpNode);
	}

	/**
	 * Returns whether the engine is currently running
	 * 
	 * @return <code>true</code> if the engine is running, <code>false</code> otherwise
	 */
	public boolean isRunning() {
		return running;
	}

	/**
	 * {@inheritDoc}
	 */
	public void init() {
		super.init();
		cache = new Cache(System.getProperty("user.home") + "/.freeMule/cache");
		cache.setEnabled(false);
		running = false;
		fcpClient = null;
	}

	/**
	 * {@inheritDoc}
	 */
	public void destroy() {
		fcpClient = null;
		cache = null;
		running = false;
		super.destroy();
	}

	/**
	 * {@inheritDoc}
	 */
	public void start() {
		FcpClientHello fcpClientHello = new FcpClientHello();
		try {
			fcpClient.execute(fcpClientHello);
			running = true;
			super.start();
		} catch (FcpException fe1) {
			fireServiceAborted(fe1);
		} catch (IOException ioe1) {
			fireServiceAborted(ioe1);
		}
	}

	/**
	 * Requests the data for the key <code>uri</code> from the node. This method follows redirects (e.g. for large SSK keys).
	 * 
	 * @param uri
	 *            The key to request
	 * @param hopsToLive
	 *            The hops-to-live value
	 * @return The data, or <code>null</code> if the data was not found
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP exception occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public Data getKey(String uri, int hopsToLive) throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		Data data = null;
		boolean isRedirect = true;
		while (isRedirect) {
			if (cache.hasKey(uri)) {
				data = cache.getKey(uri);
			} else {
				FcpClientGet fcpClientGet = new FcpClientGet(uri, hopsToLive);
				fcpClient.execute(fcpClientGet);
				data = new Data(fcpClientGet.getMetadata(), fcpClientGet.getData());
			}
			if ((data != null) && (data.getData().length == 0)) {
				StringTokenizer stok = new StringTokenizer(new String(data.getMetadata()), "\n");
				isRedirect = false;
				while (stok.hasMoreTokens()) {
					String line = stok.nextToken().trim();
					if (line.startsWith("Redirect.Target=")) {
						uri = line.substring(16).trim();
						isRedirect = true;
					}
				}
			} else {
				isRedirect = false;
			}
		}
		return data;
	}

	/**
	 * Stores the given data under the specified key.
	 * 
	 * @param uri
	 *            The key to store the data at
	 * @param hopsToLive
	 *            The hops-to-live value
	 * @param data
	 *            The data to store
	 * @return <code>true</code> if the data was stored at the key, <code>false</code> otherwise
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP exception occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public boolean putKey(String uri, int hopsToLive, Data data) throws EngineStoppedException, FcpException, IOException {
		return putKey(uri, hopsToLive, data, false);
	}

	/**
	 * Stores the given data under the specified key.
	 * 
	 * @param uri
	 *            The key to store the data at
	 * @param hopsToLive
	 *            The hops-to-live value
	 * @param data
	 *            The data to store
	 * @param swallowCollision
	 *            If <code>true</code>, doesn't throw FcpKeyCollisionException
	 * @return <code>true</code> if the data was stored at the key, <code>false</code> otherwise
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP exception occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public boolean putKey(String uri, int hopsToLive, Data data, boolean swallowCollision) throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		if ("CHK@".equals(uri)) {
			FcpClientPut fcpClientPut = new FcpClientPut("freenet:" + uri, hopsToLive, data.getData(), data.getMetadata());
			fcpClient.execute(fcpClientPut);
			cache.putKey(fcpClientPut.getFinalURI(), data);
		} else {
			if (cache.hasKey(uri)) {
				return false;
			}
			if ((uri.startsWith("SSK@") || uri.startsWith("KSK@")) && (data.getData().length > 32768)) {
				FcpClientPut fcpClientPut = new FcpClientPut("freenet:CHK@", hopsToLive, data.getData(), data.getMetadata());
				fcpClient.execute(fcpClientPut);
				FcpRedirect redirect = new FcpRedirect(fcpClientPut.getFinalURI());
				fcpClientPut = new FcpClientPut("freenet:" + uri, hopsToLive, new byte[0], redirect.toByteArray());
				try {
					fcpClient.execute(fcpClientPut);
				} catch (FcpKeyCollisionException fkce1) {
					if (!swallowCollision) {
						throw fkce1;
					}
				}
				cache.putKey(uri, data);
			} else {
				FcpClientPut fcpClientPut = new FcpClientPut("freenet:" + uri, hopsToLive, data.getData(), data.getMetadata());
				try {
					fcpClient.execute(fcpClientPut);
				} catch (FcpKeyCollisionException fkce1) {
					if (!swallowCollision) {
						throw fkce1;
					}
				}
				cache.putKey(fcpClientPut.getFinalURI().substring(8), data);
			}
		}
		return true;
	}

	/**
	 * Generates a new SVK key pair.
	 * 
	 * @return An array of String containing the private key at index 0, the public key at index 1, and the entropy at index 2
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String[] generateSecretKey() throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		FcpGenerateSVKPair fcpGenKey = new FcpGenerateSVKPair();
		fcpClient.execute(fcpGenKey);
		return new String[] { fcpGenKey.getPrivateKey(), fcpGenKey.getPublicKey(), fcpGenKey.getCryptoKey() };
	}

	/**
	 * Creates the key for the given data without storing it.
	 * 
	 * @param data
	 *            The data to create the key for
	 * @return The final URI of the data
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String createKey(Data data) throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		FcpGenerateCHK generateCHK = new FcpGenerateCHK(data.getData(), data.getMetadata());
		fcpClient.execute(generateCHK);
		return generateCHK.getURI().substring(8);
	}

	/**
	 * Creates the key for the data returned by the {@link InputStream}.
	 * 
	 * @param metadata
	 *            The metadata of the data
	 * @param dataInput
	 *            An {@link InputStream} containing the data
	 * @param dataLength
	 *            The length of the data
	 * @return The final URI of the data
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String createKey(byte[] metadata, InputStream dataInput, long dataLength) throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		FcpGenerateCHK generateCHK = new FcpGenerateCHK(dataInput, dataLength, new ByteArrayInputStream(metadata), metadata.length);
		fcpClient.execute(generateCHK);
		return generateCHK.getURI().substring(8);
	}

	/**
	 * Stores the data returned by the {@link InputStream} in Freenet.
	 * 
	 * @param uri
	 *            The URI to store the data at. Use <code>"CHK@"</code> if the URI is not known before inserting
	 * @param hopsToLive
	 *            The Hops-to-Live value for the request
	 * @param removeLocalKey
	 *            Remove key in local data store before insert if <code>true</code>
	 * @param dataInput
	 *            An {@link InputStream} containing the data
	 * @param dataLength
	 *            The length of the data
	 * @param metadata
	 *            The metadata of the data
	 * @return The final URI of the data
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String putKey(String uri, int hopsToLive, boolean removeLocalKey, InputStream dataInput, long dataLength, byte[] metadata) throws EngineStoppedException, FcpException, IOException {
		return putKey(uri, hopsToLive, removeLocalKey, dataInput, dataLength, metadata, false);
	}

	/**
	 * Stores the data returned by the {@link InputStream} in Freenet.
	 * 
	 * @param uri
	 *            The URI to store the data at. Use <code>"CHK@"</code> if the URI is not known before inserting
	 * @param hopsToLive
	 *            The Hops-to-Live value for the request
	 * @param removeLocalKey
	 *            Remove key in local data store before insert if <code>true</code>
	 * @param dataInput
	 *            An {@link InputStream} containing the data
	 * @param dataLength
	 *            The length of the data
	 * @param metadata
	 *            The metadata of the data
	 * @param swallowCollision
	 *            If <code>true</code>, no FcpKeyCollisionException will be thrown
	 * @return The final URI of the data
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String putKey(String uri, int hopsToLive, boolean removeLocalKey, InputStream dataInput, long dataLength, byte[] metadata, boolean swallowCollision) throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		FcpClientPut generateCHK = new FcpClientPut(uri, hopsToLive, removeLocalKey, dataInput, dataLength, new ByteArrayInputStream(metadata), metadata.length);
		if (swallowCollision) {
			try {
				fcpClient.execute(generateCHK);
			} catch (FcpKeyCollisionException fkce1) {
				return uri;
			}
		} else {
			fcpClient.execute(generateCHK);
		}
		return generateCHK.getFinalURI().substring(8);
	}

	/**
	 * Stores the data returned by the {@link InputStream} in Freenet.
	 * 
	 * @param uri
	 *            The URI to store the data at. Use <code>"CHK@"</code> if the URI is not known before inserting
	 * @param hopsToLive
	 *            The Hops-to-Live value for the request
	 * @param removeLocalKey
	 *            Remove key in local data store before insert if <code>true</code>
	 * @param data
	 *            The data
	 * @param metadata
	 *            The metadata of the data
	 * @return The final URI of the data
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String putKey(String uri, int hopsToLive, boolean removeLocalKey, byte[] data, byte[] metadata) throws EngineStoppedException, FcpException, IOException {
		return putKey(uri, hopsToLive, removeLocalKey, data, metadata, false);
	}

	/**
	 * Stores the data returned by the {@link InputStream} in Freenet.
	 * 
	 * @param uri
	 *            The URI to store the data at. Use <code>"CHK@"</code> if the URI is not known before inserting
	 * @param hopsToLive
	 *            The Hops-to-Live value for the request
	 * @param removeLocalKey
	 *            Remove key in local data store before insert if <code>true</code>
	 * @param data
	 *            The data
	 * @param metadata
	 *            The metadata of the data
	 * @param swallowCollision
	 *            If <code>true</code>, no FcpKeyCollisionException will be thrown
	 * @return The final URI of the data
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public String putKey(String uri, int hopsToLive, boolean removeLocalKey, byte[] data, byte[] metadata, boolean swallowCollision) throws EngineStoppedException, FcpException, IOException {
		return putKey(uri, hopsToLive, removeLocalKey, new ByteArrayInputStream(data), data.length, metadata, swallowCollision);
	}

	/**
	 * Executes a custom FCP command.
	 * 
	 * @param fcpCommand
	 *            The FCP command to execute
	 * @throws EngineStoppedException
	 *             if the engine is not running
	 * @throws FcpException
	 *             if an FCP error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public void execute(FcpCommand fcpCommand) throws EngineStoppedException, FcpException, IOException {
		if (!isRunning()) {
			throw new EngineStoppedException();
		}
		fcpClient.execute(fcpCommand);
	}

	public Document createManifest(FreenetFile freenetFile) throws EngineStoppedException, FcpException, IOException {
		SimpleXML manifestXML = new SimpleXML("file");
		manifestXML.append("name", freenetFile.getName());
		manifestXML.append("length", Hex.toHex(freenetFile.getLength()));
		SimpleXML blocklistXML = manifestXML.append("blocklist");
		for (int blockIndex = 0, blockCount = freenetFile.getBlockCount(); blockIndex < blockCount; blockIndex++) {
			String key = freenetFile.getBlockKey(blockIndex);
			if (key == null) {
				key = createKey(new byte[0], freenetFile.getBlockInputStream(blockIndex), Math.min(FreenetFile.BLOCKSIZE, freenetFile.getLength() - blockIndex * FreenetFile.BLOCKSIZE));
				if (key.startsWith("freenet:")) {
					key = key.substring(8);
				}
			}
			blocklistXML.append("block", key);
		}
		return manifestXML.getDocument();
	}

}
