/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.w3c.dom.Document;

import de.todesbaum.freemule.util.Hex;
import de.todesbaum.freemule.util.SimpleXML;
import de.todesbaum.freemule.util.XML;
import de.todesbaum.util.freenet.fcp.FcpRuntimeException;


/**
 * Collects filelist from other clients.
 * 
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id: FilelistCollectorService.java,v 1.1 2005/11/28 22:50:39 bombe
 *          Exp $
 */
public class FilelistCollectorService extends AbstractEngineService {

	public static final String NAME = "Filelist Collecting Service";

	/**
	 * The {@link List} of regiestered {@link FilelistCollectorListener}s.
	 */
	private Set<FilelistCollectorListener> filelistCollectorListeners;

	/**
	 * The queue of clients to get filelists for.
	 */
	private List<Client> clientQueue;

	/**
	 * The keys of already collected lists, to prevent fetching a list more than
	 * once.
	 */
	private List<String> collectedLists;

	/**
	 * {@inheritDoc}
	 */
	public String getName() {
		return NAME;
	}

	/**
	 * Adds a {@link FilelistCollectorListener} to the list of registered
	 * listeners.
	 * 
	 * @param filelistCollectorListener
	 *            The {@link FilelistCollectorListener} to add
	 */
	public void addFilelistCollectorListener(FilelistCollectorListener filelistCollectorListener) {
		filelistCollectorListeners.add(filelistCollectorListener);
	}

	/**
	 * Removes a {@link FilelistCollectorListener} from the list of registered
	 * listeners.
	 * 
	 * @param filelistCollectorListener
	 *            The {@link FilelistCollectorListener} to remove
	 */
	public void removeFilelistCollectorListener(FilelistCollectorListener filelistCollectorListener) {
		filelistCollectorListeners.remove(filelistCollectorListener);
	}

	/**
	 * Notifies listeners that the filelist of the remote client was requested.
	 * 
	 * @param remoteClient
	 *            The client whose filelist to get
	 * @param key
	 *            The key of the filelist
	 */
	private void fireFilelistCollectingStarted(Client remoteClient, String key) {
		for (FilelistCollectorListener filelistCollectorListener: filelistCollectorListeners) {
			filelistCollectorListener.filelistCollectingStarted(remoteClient, key);
		}
	}

	/**
	 * Notifies listeners that the request for <code>remoteClient</code>'s
	 * filelist failed.
	 * 
	 * @param remoteClient
	 *            The client whose filelist could not be retrieved
	 * @param key
	 *            The key of the filelist
	 * @param exception
	 *            The exception that prevented retrieval
	 */
	private void fireFilelistCollectingFailed(Client remoteClient, String key, Exception exception) {
		for (FilelistCollectorListener filelistCollectorListener: filelistCollectorListeners) {
			filelistCollectorListener.filelistCollectingFailed(remoteClient, key, exception);
		}
	}

	/**
	 * Notifies listeners that a filelist was sucessfully retrieved.
	 * 
	 * @param remoteClient
	 *            The client whose filelist was retrieved
	 * @param key
	 *            The key of the filelist
	 * @param sharedFiles
	 *            The collected filelist
	 */
	private void fireFilelistCollectingSucceeded(Client remoteClient, String key, SharedFile[] sharedFiles) {
		for (FilelistCollectorListener filelistCollectorListener: filelistCollectorListeners) {
			filelistCollectorListener.filelistCollectingComplete(remoteClient, key, sharedFiles);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void init() {
		super.init();
		filelistCollectorListeners = new HashSet<FilelistCollectorListener>();
		clientQueue = new ArrayList<Client>();
		collectedLists = new ArrayList<String>();
	}

	/**
	 * {@inheritDoc}
	 */
	public void destroy() {
		filelistCollectorListeners.clear();
		filelistCollectorListeners = null;
		clientQueue.clear();
		clientQueue = null;
		collectedLists.clear();
		collectedLists = null;
		super.destroy();
	}

	/**
	 * Adds a client to the queue so that his filelist can be retrieved.
	 * 
	 * @param client
	 *            The client whose filelist to retrieve
	 */
	public void addClient(Client client) {
		synchronized (syncObject) {
			if (!clientQueue.contains(client)) {
				clientQueue.add(client);
			} else {
				clientQueue.get(clientQueue.indexOf(client)).setFilelistIndex(client.getFilelistIndex());
			}
			syncObject.notify();
		}
	}

	/**
	 * Checks the {@link #clientQueue} for clients whose filelists to get.
	 */
	protected void work() throws Throwable {
		Client remoteClient = getNextClient();
		while (!stopping && (remoteClient != null)) {
			int fetchedIndex = remoteClient.getFilelistIndex();
			String filelistKey = remoteClient.getAddress() + "/freeMule/filelist/" + Hex.toHex(fetchedIndex);
			if (!collectedLists.contains(filelistKey)) {
				try {
					fireFilelistCollectingStarted(remoteClient, filelistKey);
					Data filelistData = engine.getKey(filelistKey, hopsToLive);
					Document filelistXML = XML.transformToDocument(filelistData.getData());
					SharedFile[] sharedFiles = parseFilelist(filelistXML);
					collectedLists.add(filelistKey);
					if (remoteClient.getFilelistIndex() == fetchedIndex) {
						synchronized (clientQueue) {
							clientQueue.remove(remoteClient);
						}
					}
					fireFilelistCollectingSucceeded(remoteClient, filelistKey, sharedFiles);
				} catch (FcpRuntimeException fre1) {
					fireFilelistCollectingFailed(remoteClient, filelistKey, fre1);
				} catch (IOException ioe1) {
					fireFilelistCollectingFailed(remoteClient, filelistKey, ioe1);
				}
			}
			// check for next loop
			remoteClient = getNextClient();
		}
	}

	private SharedFile[] parseFilelist(Document filelistDocument) {
		SimpleXML filelistXML = SimpleXML.fromDocument(filelistDocument);
		List<SharedFile> resultList = new ArrayList<SharedFile>();
		SimpleXML[] fileXMLs = filelistXML.getNode("files").getNodes("file");
		for (SimpleXML fileXML: fileXMLs) {
			SharedFile sharedFile = new SharedFile();
			sharedFile.setKey(fileXML.getNode("key").getValue());
			sharedFile.setName(fileXML.getNode("name").getValue());
			sharedFile.setLength(Hex.toLong(fileXML.getNode("length").getValue()));
			sharedFile.setComment(fileXML.getNode("comment").getValue());
			sharedFile.setRating((int) Hex.toLong(fileXML.getNode("rating").getValue()));
			resultList.add(sharedFile);
		}
		return resultList.toArray(new SharedFile[resultList.size()]);
	}

	/**
	 * Gets the next client from the client queue.
	 * 
	 * @return The next client, or <code>null</code> if there is no next
	 *         client
	 */
	private Client getNextClient() {
		synchronized (syncObject) {
			if (clientQueue.size() > 0) {
				return clientQueue.get((int) (Math.random() * clientQueue.size()));
			}
		}
		return null;
	}

}
