/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;

import de.todesbaum.freemule.util.Hex;
import de.todesbaum.freemule.util.SimpleXML;
import de.todesbaum.freemule.util.XML;
import de.todesbaum.util.freenet.fcp.FcpException;
import de.todesbaum.util.freenet.fcp.FcpRuntimeException;


/**
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:FilelistInsertService.java 220 2006-03-06 14:46:23Z bombe $
 */
public class FilelistInsertService extends AbstractEngineService {

	public static final String NAME = "Filelist Insert Service";

	private List<FilelistInsertListener> filelistInsertListeners;

	private List<SharedFile[]> sharedFiles = null;
	private List<Integer> indexQueue = null;
	private String clientPrivateAddress;
	private FreenetFile[] downloadFiles = null;

	public String getName() {
		return NAME;
	}

	public void addFilelistInsertListener(FilelistInsertListener filelistInsertListener) {
		if (!filelistInsertListeners.contains(filelistInsertListener)) {
			filelistInsertListeners.add(filelistInsertListener);
		}
	}

	public void removeFilelistInsertListener(FilelistInsertListener filelistInsertListener) {
		filelistInsertListeners.remove(filelistInsertListener);
	}

	/**
	 * Notifies listeners that a filelist insert is about to start.
	 * 
	 * @param key
	 *            The key of the filelist
	 * @param currentIndex
	 *            The current index of the filelist
	 */
	private void fireFilelistInsertStarted(String key, int currentIndex) {
		for (FilelistInsertListener filelistInsertListener: filelistInsertListeners) {
			filelistInsertListener.filelistInsertStarted(key, currentIndex);
		}
	}

	/**
	 * Notifies listeners that a filelist insert failed.
	 * 
	 * @param key
	 *            The key of the filelist
	 * @param currentIndex
	 *            The current index of the filelist
	 * @param exception
	 *            The exception that prevented insertion
	 */
	private void fireFilelistInsertFailed(String key, int currentIndex, Exception exception) {
		for (FilelistInsertListener filelistInsertListener: filelistInsertListeners) {
			filelistInsertListener.filelistInsertFailed(key, currentIndex, exception);
		}
	}

	/**
	 * Notifies listeners that the filelist was successfully inserted.
	 * 
	 * @param key
	 *            The key of the filelist
	 * @param index
	 *            The current index of the filelist
	 */
	private void fireFilelistInsertComplete(String key, int index) {
		for (FilelistInsertListener filelistInsertListener: filelistInsertListeners) {
			filelistInsertListener.filelistInsertComplete(key, index);
		}
	}

	public void setClientPrivateAddress(String clientPrivateAddress) {
		this.clientPrivateAddress = clientPrivateAddress;
	}

	public void addSharedFiles(int index, SharedFile[] files) {
		synchronized (syncObject) {
			this.sharedFiles.add(files);
			this.indexQueue.add(index);
			syncObject.notify();
		}
	}

	public void setDownloadFiles(FreenetFile[] downloadFiles) {
		synchronized (syncObject) {
			this.downloadFiles = new FreenetFile[downloadFiles.length];
			System.arraycopy(downloadFiles, 0, this.downloadFiles, 0, downloadFiles.length);
			syncObject.notify();
		}
	}

	public void init() {
		super.init();
		filelistInsertListeners = new ArrayList<FilelistInsertListener>();
		sharedFiles = new ArrayList<SharedFile[]>();
		indexQueue = new ArrayList<Integer>();
		clientPrivateAddress = null;
	}

	public void start() {
		super.start();
	}

	public void destroy() {
		clientPrivateAddress = null;
		filelistInsertListeners.clear();
		filelistInsertListeners = null;
		sharedFiles.clear();
		sharedFiles = null;
		indexQueue.clear();
		indexQueue = null;
		super.destroy();
	}

	protected void work() {
		String keyPrefix = clientPrivateAddress + "/freeMule/filelist/";
		while (!stopping) {
			Document filelist;
			int index;
			synchronized (syncObject) {
				if (sharedFiles.size() == 0) {
					break;
				}
				SharedFile[] files = sharedFiles.get(0);
				index = indexQueue.get(0);
				filelist = createFilelist(files);
			}
			byte[] filelistContent = XML.transformToByteArray(filelist);
			Data filelistData = new Data(new byte[0], filelistContent);

			String key = keyPrefix + Hex.toHex(index);
			try {
				fireFilelistInsertStarted(key, index);
				engine.putKey(key, hopsToLive, filelistData, true);
				synchronized (syncObject) {
					sharedFiles.remove(0);
					indexQueue.remove(0);
				}
				fireFilelistInsertComplete(key, index);
			} catch (FcpRuntimeException fre1) {
				fireFilelistInsertFailed(key, index, fre1);
			} catch (IOException ioe1) {
				fireFilelistInsertFailed(key, index, ioe1);
			} catch (FcpException fe1) {
				fireServiceAborted(fe1);
				stopping = true;
				break;
			} catch (EngineStoppedException ese1) {
				fireServiceAborted(ese1);
				stopping = true;
				break;
			}
		}
	}

	private Document createFilelist(SharedFile[] files) {
		SimpleXML filelistXML = new SimpleXML("filelist");
		SimpleXML filesXML = filelistXML.append("files");
		for (SharedFile sharedFile: files) {
			SimpleXML fileXML = filesXML.append("file");
			fileXML.append("name", sharedFile.getName());
			fileXML.append("length", Hex.toHex(sharedFile.getLength()));
			fileXML.append("key", sharedFile.getKey());
			fileXML.append("comment", sharedFile.getComment());
			fileXML.append("rating", Hex.toHex(sharedFile.getRating()));
		}
		return filelistXML.getDocument();
	}

}
