/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.BitSet;

import de.todesbaum.freemule.io.FilePartInputStream;
import de.todesbaum.freemule.io.FilePartOutputStream;
import de.todesbaum.freemule.util.Hex;
import de.todesbaum.freemule.util.SimpleXML;


/**
 * An extension of {@link FreenetFile} that supports segmentation and check
 * blocks.
 * 
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:FreenetFile.java 220 2006-03-06 14:46:23Z bombe $
 */
public class FreenetFile {

	public static final int BLOCKSIZE = 1048576;

	/**
	 * The name of the file.
	 */
	protected String name;

	/**
	 * The key of the file, or <code>null</code> if the key is not yet known.
	 */
	protected String key;

	/**
	 * The length of the file, or <code>-1</code> if the length is not yet
	 * known.
	 */
	protected long length = -1;

	/**
	 * The name of the local file that stores the data of the FreenetFile.
	 */
	protected String dataFilename;

	/**
	 * The name of the file that stores the properties.
	 */
	protected String propertiesFilename;

	/**
	 * The keys for the data blocks.
	 */
	protected String[] blockKeys;

	/**
	 * The status for the manifest.
	 */
	protected BlockStatus manifestStatus;

	/**
	 * The status for the data blocks.
	 */
	protected BlockStatus[] blockStatus;

	/**
	 * The status of the file.
	 */
	protected FileStatus status;

	public FreenetFile() {
		super();
	}

	/**
	 * Creates a new {@link FreenetFile} with the same properties as
	 * <code>freenetFile</code>.
	 * 
	 * @param freenetFile
	 *            The file to copy
	 */
	public FreenetFile(FreenetFile freenetFile) {
		name = freenetFile.name;
		key = freenetFile.key;
		dataFilename = freenetFile.dataFilename;
		status = freenetFile.status;
		manifestStatus = freenetFile.manifestStatus;
		setLength(freenetFile.length);
		int blockCount = getBlockCount();
		blockKeys = new String[blockCount];
		blockStatus = new BlockStatus[blockCount];
		System.arraycopy(freenetFile.blockKeys, 0, blockKeys, 0, blockCount);
		System.arraycopy(freenetFile.blockStatus, 0, blockStatus, 0, blockCount);
	}

	/**
	 * Returns the key of the file, or <code>null</code> if the key is not yet
	 * known.
	 * 
	 * @return The key of the file, or <code>null</code>
	 */
	public synchronized String getKey() {
		return key;
	}

	/**
	 * Sets the key of the file. <code>key</code> can be null to signal that
	 * the key is not yet known.
	 * 
	 * @param key
	 *            The key to set, or <code>null</code>
	 */
	public synchronized void setKey(String key) {
		this.key = key;
	}

	/**
	 * Returns the length of the file, or <code>-1</code> if the length is not
	 * yet known.
	 * 
	 * @return The length of the file, or <code>-1</code>
	 */
	public synchronized long getLength() {
		return length;
	}

	/**
	 * Sets the length of the file. <code>length</code> can be <code>-1</code>
	 * to signal that the length is now yet known.
	 * 
	 * @param length
	 *            The length to set, or <code>-1</code>
	 */
	public synchronized void setLength(long length) {
		this.length = length;
		int blockCount = getBlockCount();
		blockKeys = new String[blockCount];
		blockStatus = new BlockStatus[blockCount];
		for (int blockIndex = 0; blockIndex < blockCount; blockIndex++) {
			blockKeys[blockIndex] = null;
			blockStatus[blockIndex] = BlockStatus.notAvailable;
		}
	}

	/**
	 * Returns the name of the file.
	 * 
	 * @return The name of the file
	 */
	public synchronized String getName() {
		return name;
	}

	/**
	 * Sets the name of the file.
	 * 
	 * @param name
	 *            The name to set
	 */
	public synchronized void setName(String name) {
		this.name = name;
	}

	/**
	 * Returns the name of the file that holds the data of this file.
	 * 
	 * @return The name of the file that holds the data of this file
	 */
	public synchronized String getDataFilename() {
		return dataFilename;
	}

	/**
	 * Sets the name of the file that holds the data of this file.
	 * 
	 * @param dataFilename
	 *            The name of the file that holds the data of this file
	 */
	public synchronized void setDataFilename(String dataFilename) {
		this.dataFilename = dataFilename;
	}

	/**
	 * Returns an {@link InputStream} that returns the data of the file.
	 * 
	 * @return An {@link InputStream} that returns the data of the file
	 * @throws FileNotFoundException
	 *             if the data file does not exist
	 */
	public synchronized InputStream getDataInputStream() throws FileNotFoundException {
		return new FileInputStream(new File(dataFilename));
	}

	/**
	 * Returns an {@link OutputStream} that writes to the data file.
	 * <p>
	 * <strong>Warning:</strong> the data file will be overwritten!
	 * 
	 * @return An {@link OutputStream} that writes to the data file.
	 * @throws FileNotFoundException
	 *             if the data file can not be created
	 */
	public synchronized OutputStream getDataOutputStream() throws FileNotFoundException {
		return new FileOutputStream(new File(dataFilename));
	}

	/**
	 * Sets the {@link FileStatus} of this file.
	 * 
	 * @param fileStatus
	 *            The {@link FileStatus} to set
	 */
	public synchronized void setStatus(FileStatus fileStatus) {
		this.status = fileStatus;
	}

	/**
	 * Returns the {@link FileStatus} of this file.
	 * 
	 * @return The {@link FileStatus} of this file
	 */
	public synchronized FileStatus getStatus() {
		return status;
	}

	/**
	 * @return Returns the manifestStatus.
	 */
	public synchronized BlockStatus getManifestStatus() {
		return manifestStatus;
	}

	/**
	 * @param manifestStatus
	 *            The manifestStatus to set.
	 */
	public synchronized void setManifestStatus(BlockStatus manifestStatus) {
		this.manifestStatus = manifestStatus;
	}

	/**
	 * Returns the filename of the properties file.
	 * 
	 * @return The filename of the properties file
	 */
	public String getPropertiesFilename() {
		return propertiesFilename;
	}

	/**
	 * Sets the filename of the properties file.
	 * 
	 * @param propertiesFilename
	 *            The filename of the properties file
	 */
	public void setPropertiesFilename(String propertiesFilename) {
		this.propertiesFilename = propertiesFilename;
	}

	public void setBlockKey(int blockNumber, String key) {
		blockKeys[blockNumber] = key;
	}

	public String getBlockKey(int blockNumber) {
		return blockKeys[blockNumber];
	}

	public void setBlockStatus(int blockIndex, BlockStatus blockStatus) {
		this.blockStatus[blockIndex] = blockStatus;
	}

	public BlockStatus getBlockStatus(int blockIndex) {
		return blockStatus[blockIndex];
	}

	public BitSet getBlockList() {
		return getBlockList(BlockStatus.available);
	}

	public BitSet getBlockList(BlockStatus desiredStatus) {
		BitSet resultBitSet = new BitSet(getBlockCount());
		for (int blockIndex = 0, blockCount = getBlockCount(); blockIndex < blockCount; blockIndex++) {
			if (blockStatus[blockIndex] == desiredStatus) {
				resultBitSet.set(blockIndex);
			}
		}
		return resultBitSet;
	}

	/**
	 * Returns an {@link InputStream} that streams a single data block.
	 * 
	 * @param blockNumber
	 *            The number of the block
	 * @return An {@link InputStream} that streams the specified block
	 * @throws FileNotFoundException
	 *             if {@link #dataFilename} can not be found
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public synchronized InputStream getBlockInputStream(int blockNumber) throws FileNotFoundException, IOException {
		long blockOffset = blockNumber * BLOCKSIZE;
		return new FilePartInputStream(dataFilename, blockOffset, Math.min(BLOCKSIZE, length - blockOffset));
	}

	/**
	 * Returns an {@link OutputStream} that streams to position in the data file
	 * that belongs to the specified block.
	 * 
	 * @param blockNumber
	 *            The number of the block
	 * @return An {@link OutputStream} that will dump the data to the correct
	 *         place on the disk
	 * @throws FileNotFoundException
	 *             if {@link #dataFilename} can not be created
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	public synchronized OutputStream getBlockOutputStream(int blockNumber) throws FileNotFoundException, IOException {
		long blockOffset = blockNumber * BLOCKSIZE;
		return new FilePartOutputStream(dataFilename, blockOffset, Math.min(BLOCKSIZE, length - blockOffset));
	}

	/**
	 * Checks whether <code>o</code> is a FreenetFile and has the same key as
	 * this FreenetFile.
	 * 
	 * @param o
	 *            The object to check for equality
	 * @return <code>true</code> if <code>o</code> is a FreenetFile and has
	 *         the same key as this FreenetFie
	 */
	@Override
	public boolean equals(Object o) {
		if ((o == null) || !(o instanceof FreenetFile)) {
			return false;
		}
		FreenetFile freenetFile = (FreenetFile) o;
		return freenetFile.key.equals(key);
	}

	/**
	 * Returns the hash code of this FreenetFile's key.
	 * 
	 * @see String#hashCode()
	 * @return The hash code of this FreenetFile's key
	 */
	@Override
	public int hashCode() {
		return key.hashCode();
	}

	public SimpleXML getXML(String elementName) {
		SimpleXML documentXML = new SimpleXML(elementName);
		documentXML.append("name", name);
		documentXML.append("length", Hex.toHex(length));
		documentXML.append("key", key);
		documentXML.append("status", Hex.toHex(status.ordinal()));
		documentXML.append("manifest-status", Hex.toHex(manifestStatus.ordinal()));
		documentXML.append("datafile-name", dataFilename);
		documentXML.append("propertiesfile-name", propertiesFilename);
		SimpleXML blockListXML = documentXML.append("block-list");
		for (int blockIndex = 0, blockCount = getBlockCount(); blockIndex < blockCount; blockIndex++) {
			SimpleXML blockXML = blockListXML.append("block");
			blockXML.append("block-status", Hex.toHex(blockStatus[blockIndex].ordinal()));
			blockXML.append("block-key", blockKeys[blockIndex]);
		}
		return documentXML;
	}

	public static FreenetFile createFromXML(SimpleXML documentXML) {
		FreenetFile freenetFile = new FreenetFile();

		freenetFile.setName(documentXML.getNode("name").getValue());
		freenetFile.setKey(documentXML.getNode("key").getValue());
		freenetFile.setLength(Hex.toLong(documentXML.getNode("length").getValue()));
		freenetFile.setStatus(FileStatus.values()[(int) Hex.toLong(documentXML.getNode("status").getValue())]);
		freenetFile.setManifestStatus(BlockStatus.values()[(int) Hex.toLong(documentXML.getNode("manifest-status").getValue())]);
		freenetFile.setDataFilename(documentXML.getNode("datafile-name").getValue());
		freenetFile.setPropertiesFilename(documentXML.getNode("propertiesfile-name").getValue());
		SimpleXML[] blockXMLs = documentXML.getNode("block-list").getNodes("block");
		int blockIndex = 0;
		for (SimpleXML blockXML: blockXMLs) {
			freenetFile.setBlockStatus(blockIndex, BlockStatus.values()[(int) Hex.toLong(blockXML.getNode("block-status").getValue())]);
			freenetFile.setBlockKey(blockIndex, blockXML.getNode("block-key").getValue());
			blockIndex++;
		}

		return freenetFile;
	}

	/**
	 * @return Returns the blockCount.
	 */
	public int getBlockCount() {
		return (int) (((length - 1) / BLOCKSIZE) + 1);
	}

}
