/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import de.todesbaum.freemule.io.StreamCopier;
import de.todesbaum.freemule.util.SimpleXML;
import de.todesbaum.freemule.util.XML;
import de.todesbaum.util.freenet.fcp.FcpException;


/**
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:SharedFileService.java 220 2006-03-06 14:46:23Z bombe $
 */
public class SharedFileService extends AbstractEngineService {

	public static final String NAME = "Shared File Service";

	private SharedDirectory[] sharedDirectories;
	private List<SharedFile> sharedFiles;
	private List<SharedFileListener> sharedFileListeners;
	private boolean changed = false;

	public String getName() {
		return NAME;
	}

	public void addSharedFileListener(SharedFileListener sharedFileListener) {
		if (!sharedFileListeners.contains(sharedFileListener)) {
			sharedFileListeners.add(sharedFileListener);
		}
	}

	public void removeSharedFileListener(SharedFileListener sharedFileListener) {
		sharedFileListeners.remove(sharedFileListener);
	}

	/**
	 * Notifies listeners that the loading of the list of shared files is about
	 * to start.
	 * 
	 * @param filename
	 *            The name of the file containing the list of shared files
	 */
	private void fireSharedFileLoadingStarted(String filename) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileLoadingStarted(filename);
		}
	}

	/**
	 * Notifies listeners that the loading of the list of shared files failed.
	 * 
	 * @param filename
	 *            The name of the file containing the list of shared files
	 * @param exception
	 *            The exception that prevented loading
	 */
	private void fireSharedFileLoadingFailed(String filename, Exception exception) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileLoadingFailed(filename, exception);
		}
	}

	/**
	 * Notifies listeners that the loading of the list of shared files is
	 * complete.
	 * 
	 * @param filename
	 *            The name of the file containing the list of shared files.
	 */
	private void fireSharedFileLoadingComplete(String filename) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileLoadingComplete(filename);
		}
	}

	/**
	 * Notifies listeners that a previously shared directory is not shared
	 * anymore.
	 * 
	 * @param directory
	 *            The directory that is no longer shared
	 */
	private void fireSharedFileUnsharedDirectory(String directory) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileUnsharedDirectory(directory);
		}
	}

	/**
	 * Notifies listeners that a previously shared file is not shared anymore.
	 * 
	 * @param filename
	 *            The name of the file that is no longer shared
	 */
	private void fireSharedFileUnsharedFile(String filename) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileUnsharedFile(filename);
		}
	}

	/**
	 * Notifies listeners that the list of shared files is about to being saved.
	 * 
	 * @param filename
	 *            The name of the file the list of shared files is written to
	 */
	private void fireSharedFileSavingStarted(String filename) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileSavingStarted(filename);
		}
	}

	/**
	 * Notifies listeners that the saving of the list of shared files failed.
	 * 
	 * @param filename
	 *            The name of the file the list of shared files is being written
	 *            to
	 * @param exception
	 *            The exception that prevented saving
	 */
	private void fireSharedFileSavingFailed(String filename, Exception exception) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileSavingFailed(filename, exception);
		}
	}

	/**
	 * Notifies listeners that the saving of the list of shared files is
	 * complete.
	 * 
	 * @param filename
	 *            The name of the file the list of shared files was written to
	 */
	private void fireSharedFileSavingComplete(String filename) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileSavingComplete(filename);
		}
	}

	/**
	 * Notifies listeners that a scan of all shared directories is about to
	 * start.
	 */
	private void fireSharedFileScanStarted() {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileScanStarted();
		}
	}

	/**
	 * Notifies listeners that the scan of all shared directories has been
	 * completed.
	 */
	private void fireSharedFileScanComplete() {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileScanComplete();
		}
	}

	/**
	 * Notifies listeners that the scan of a shared directory is about to start.
	 * 
	 * @param directory
	 *            The shared directory that will be scanned
	 */
	private void fireSharedFileDirectoryScanStarted(String directory) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileDirectoryScanStarted(directory);
		}
	}

	/**
	 * Notifies listeners that the scan of a shared directory failed.
	 * 
	 * @param directory
	 *            The shared directory that was scanned
	 * @param exception
	 *            The exception that prevented scanning
	 */
	private void fireSharedFileDirectoryScanFailed(String directory, Exception exception) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileDirectoryScanFailed(directory, exception);
		}
	}

	/**
	 * Notifies listeners that the scan of a shared directory is complete.
	 * 
	 * @param directory
	 *            The shared directory that was scanned
	 */
	private void fireSharedFileDirectoryScanComplete(String directory) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileDirectoryScanComplete(directory);
		}
	}

	/**
	 * Notifies listeners that the hashing of a file is about to start
	 * 
	 * @param filename
	 *            The name of the file that is being hashed
	 */
	private void fireSharedFileHashingStarted(String filename) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileHashingStarted(filename);
		}
	}

	/**
	 * Notifies listeners that the hashing of a shared file failed.
	 * 
	 * @param filename
	 *            The name of the file that was hashed
	 * @param exception
	 *            The exception that prevented hashing
	 */
	private void fireSharedFileHashingFailed(String filename, Exception exception) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileHashingFailed(filename, exception);
		}
	}

	private void fireSharedFileHashingProgress(String filename, int blockIndex, int blockCount) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileHashingProgress(filename, blockIndex, blockCount);
		}
	}

	/**
	 * Notifies listeners that the hashing of a file was completed.
	 * 
	 * @param filename
	 *            The file that was hashed
	 * @param key
	 *            The resulting key
	 */
	private void fireSharedFileHashingComplete(String filename, String key, long filelength, long duration) {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileHashingComplete(filename, key, filelength, duration);
		}
	}

	/**
	 * Notifies listeners that the file list has changed.
	 */
	private void fireSharedFileListChanged() {
		for (SharedFileListener sharedFileListener: sharedFileListeners) {
			sharedFileListener.sharedFileListChanged();
		}
	}

	public void setSharedDirectories(SharedDirectory[] sharedDirectories) {
		this.sharedDirectories = new SharedDirectory[sharedDirectories.length];
		System.arraycopy(sharedDirectories, 0, this.sharedDirectories, 0, sharedDirectories.length);
		readListOfSharedFiles();
	}

	/**
	 * Returns an array containing all currently shared files.
	 * 
	 * @return All currently shared files
	 */
	public SharedFile[] getSharedFiles() {
		return sharedFiles.toArray(new SharedFile[sharedFiles.size()]);
	}

	public void setChanged(boolean changed) {
		this.changed = changed;
	}

	public boolean isChanged() {
		return changed;
	}

	public void init() {
		super.init();
		sharedFileListeners = new ArrayList<SharedFileListener>();
		sharedFiles = new ArrayList<SharedFile>();
		sharedDirectories = null;
	}

	public void start() {
		super.start();
	}

	public void stop() {
		super.stop();
		saveListOfSharedFiles();
	}

	public void destroy() {
		sharedFileListeners.clear();
		sharedFileListeners = null;
		sharedFiles.clear();
		sharedFiles = null;
		sharedDirectories = null;
		super.destroy();
	}

	protected void work() {
		fireSharedFileScanStarted();
		for (SharedDirectory sharedDirectory: sharedDirectories) {
			if (stopping) {
				break;
			}
			File directory = new File(sharedDirectory.getDirectory());
			fireSharedFileDirectoryScanStarted(sharedDirectory.getDirectory());
			scanDirectory(sharedDirectory, directory, sharedDirectory.isRecursive());
			fireSharedFileDirectoryScanComplete(sharedDirectory.getDirectory());
		}
		fireSharedFileScanComplete();
		if (changed) {
			fireSharedFileListChanged();
			changed = false;
		}
	}

	private void scanDirectory(SharedDirectory sharedDirectory, File directory, final boolean recursive) {
		File[] files = directory.listFiles(new FileFilter() {

			public boolean accept(File pathname) {
				return recursive || pathname.isFile();
			}
		});
		if (files == null) {
			fireSharedFileDirectoryScanFailed(sharedDirectory.getDirectory(), new IOException("Can not read directory"));
		}
		for (File file: files) {
			if (stopping) {
				break;
			}
			if (file.isDirectory()) {
				scanDirectory(sharedDirectory, file, recursive);
				continue;
			}
			String fileName = file.getAbsolutePath();
			fileName = fileName.substring(sharedDirectory.getDirectory().length());
			if (fileName.startsWith(File.separator)) {
				fileName = fileName.substring(1);
			}
			long fileTimestamp = file.lastModified();
			// now check if we already know this file
			SharedFile sharedFile = getSharedFile(sharedDirectory.getDirectory(), fileName);
			if (sharedFile != null) {
				if (file.length() == 0) {
					fireSharedFileUnsharedFile(sharedFile.getDataFilename());
					changed = true;
					saveListOfSharedFiles();
					continue;
				}
				if (fileTimestamp > sharedFile.getTimestamp()) {
					// rehash!
					fireSharedFileHashingStarted(file.getAbsolutePath());
					long before = System.currentTimeMillis();
					FreenetFile freenetFile = createFreenetFile(file, fileName);
					long after = System.currentTimeMillis();
					sharedFile.setKey(freenetFile.getKey());
					sharedFile.setTimestamp(fileTimestamp);
					sharedFile.setDataFilename(file.getPath());
					fireSharedFileHashingComplete(file.getAbsolutePath(), freenetFile.getKey(), file.length(), after - before);
					changed = true;
					saveListOfSharedFiles();
				}
			} else {
				if (file.length() == 0) {
					continue;
				}
				fireSharedFileHashingStarted(file.getAbsolutePath());
				long before = System.currentTimeMillis();
				FreenetFile freenetFile = createFreenetFile(file, fileName);
				long after = System.currentTimeMillis();
				if (freenetFile != null) {
					sharedFile = new SharedFile(freenetFile);
					sharedFile.setTimestamp(file.lastModified());
					sharedFile.setDirectory(sharedDirectory.getDirectory());
					sharedFile.setComment("");
					sharedFile.setDataFilename(file.getPath());
					sharedFiles.add(sharedFile);
					fireSharedFileHashingComplete(file.getAbsolutePath(), freenetFile.getKey(), file.length(), after - before);
					changed = true;
					saveListOfSharedFiles();
				}
			}
		}
	}

	/**
	 * @param directory
	 * @param fileName
	 * @return The shared file, or null if there is no such file
	 */
	private SharedFile getSharedFile(String directory, String fileName) {
		for (SharedFile sharedFile: sharedFiles) {
			if (sharedFile.getDirectory().equals(directory) && sharedFile.getName().equals(fileName)) {
				return sharedFile;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link SharedFile shared file} with the specified key.
	 * 
	 * @param key
	 *            The key to search
	 * @return The {@link SharedFile shared file} with the specified key, or
	 *         <code>null</code> if we don't have a file with this key
	 */
	public SharedFile getSharedFile(String key) {
		if (key.startsWith("freenet:")) {
			key = key.substring(8);
		}
		for (SharedFile sharedFile: sharedFiles) {
			if (sharedFile.getKey().equals(key)) {
				return sharedFile;
			}
		}
		return null;
	}

	private FreenetFile createFreenetFile(File file, String filename) {
		try {
			FreenetFile freenetFile = new FreenetFile();
			freenetFile.setLength(file.length());
			freenetFile.setDataFilename(file.getAbsolutePath());
			freenetFile.setName(filename);
			freenetFile.setStatus(FileStatus.complete);
			freenetFile.setManifestStatus(BlockStatus.available);
			int blockCount = freenetFile.getBlockCount();
			for (int blockIndex = 0; blockIndex < blockCount; blockIndex++) {
				InputStream blockInput = freenetFile.getBlockInputStream(blockIndex);
				String key = engine.createKey(new byte[0], blockInput, Math.min(FreenetFile.BLOCKSIZE, freenetFile.getLength() - (blockIndex * FreenetFile.BLOCKSIZE)));
				freenetFile.setBlockKey(blockIndex, key);
				freenetFile.setBlockStatus(blockIndex, BlockStatus.available);
				fireSharedFileHashingProgress(filename, blockIndex, blockCount);
			}
			byte[] fileMetadata = XML.transformToByteArray(freenetFile.getXML("file").getDocument());
			String fileKey = engine.createKey(new Data(fileMetadata, new byte[0]));
			freenetFile.setKey(fileKey);
			return freenetFile;
		} catch (IOException ioe1) {
			fireSharedFileHashingFailed(file.getAbsolutePath(), ioe1);
		} catch (FcpException fcpe1) {
			fireServiceAborted(fcpe1);
			stopping = true;
		} catch (EngineStoppedException ese1) {
			fireServiceAborted(ese1);
			stopping = true;
		}
		return null;
	}

	private void readListOfSharedFiles() {
		File configDirectory = new File(System.getProperty("user.home") + "/.freeMule");
		File sharedFilesFile = new File(configDirectory, "shared");
		fireSharedFileLoadingStarted(sharedFilesFile.getAbsolutePath());
		if (!configDirectory.exists() || !configDirectory.isDirectory()) {
			fireSharedFileLoadingFailed(sharedFilesFile.getAbsolutePath(), new IOException("The directory does not exist."));
			return;
		}
		sharedFiles.clear();
		if (!sharedFilesFile.exists()) {
			fireSharedFileLoadingComplete(sharedFilesFile.getAbsolutePath());
			return;
		}
		ByteArrayOutputStream xmlOutput = null;
		FileInputStream sharedFilesInput = null;
		SimpleXML sharedFilesXML = null;
		try {
			xmlOutput = new ByteArrayOutputStream((int) sharedFilesFile.length());
			sharedFilesInput = new FileInputStream(sharedFilesFile);
			StreamCopier.copy(sharedFilesInput, xmlOutput, sharedFilesFile.length());
			xmlOutput.close();
			sharedFilesInput.close();
			sharedFilesXML = SimpleXML.fromDocument(XML.transformToDocument(xmlOutput.toByteArray()));
		} catch (IOException ioe1) {
			fireSharedFileLoadingFailed(sharedFilesFile.getAbsolutePath(), ioe1);
			return;
		} finally {
			if (xmlOutput != null)
				try {
					xmlOutput.close();
				} catch (IOException ioe1) {
				}
			if (sharedFilesInput != null)
				try {
					sharedFilesInput.close();
				} catch (IOException ioe1) {
				}
		}
		if (sharedFilesXML == null) {
			fireSharedFileLoadingComplete(sharedFilesFile.getAbsolutePath());
			return;
		}
		sharedFiles = new ArrayList<SharedFile>();
		SimpleXML[] sharedFileNodesXML = sharedFilesXML.getNodes("shared-file");
		for (SimpleXML sharedFileXML: sharedFileNodesXML) {
			try { // big, bad try-catch-loop. i'm lazy.
				String fileDirectory = sharedFileXML.getNode("directory").getValue();
				boolean found = false;
				String canonicalDirectory = new File(fileDirectory).getCanonicalPath();
				for (SharedDirectory sharedDirectory: sharedDirectories) {
					String canonicalSharedDirectory = new File(sharedDirectory.getDirectory()).getCanonicalPath();
					if (canonicalDirectory.equals(canonicalSharedDirectory)) {
						found = true;
						break;
					}
				}
				if (!found) {
					fireSharedFileUnsharedDirectory(fileDirectory);
					// not sharing anymore
					changed = true;
					continue;
				}
				String filePath = sharedFileXML.getNode("name").getValue();
				if (!new File(fileDirectory, filePath).exists()) {
					fireSharedFileUnsharedFile(filePath);
					changed = true;
					continue;
				}
				SharedFile sharedFile = SharedFile.createFromXML(sharedFileXML);
				sharedFile.setDataFilename(new File(fileDirectory, filePath).getPath());
				sharedFiles.add(sharedFile);
			} catch (IOException ioe1) {
				fireSharedFileLoadingFailed(sharedFilesFile.getAbsolutePath(), ioe1);
			}
		}
		fireSharedFileLoadingComplete(sharedFilesFile.getAbsolutePath());
	}

	public void saveListOfSharedFiles() {
		File configDirectory = new File(System.getProperty("user.home"), ".freeMule");
		File configFile = new File(configDirectory, "shared");
		fireSharedFileSavingStarted(configFile.getAbsolutePath());
		if (!configDirectory.exists()) {
			configDirectory.mkdirs();
		}

		SimpleXML sharedFilesXML = new SimpleXML("shared-files");
		for (SharedFile sharedFile: sharedFiles) {
			sharedFilesXML.append(sharedFile.getXML("shared-file"));
		}

		FileOutputStream configFileOutput = null;
		try {
			configFileOutput = new FileOutputStream(configFile);
			configFileOutput.write(XML.transformToByteArray(sharedFilesXML.getDocument()));
		} catch (IOException ioe1) {
			fireSharedFileSavingFailed(configFile.getAbsolutePath(), ioe1);
		} finally {
			if (configFileOutput != null) {
				try {
					configFileOutput.close();
				} catch (IOException ioe1) {
				}
			}
		}
		fireSharedFileSavingComplete(configFile.getAbsolutePath());
	}
}
