/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Calendar;
import java.util.List;
import java.util.TimeZone;

import de.todesbaum.freemule.util.DateFormatter;
import de.todesbaum.freemule.util.Hex;
import de.todesbaum.freemule.util.SimpleXML;
import de.todesbaum.freemule.util.XML;
import de.todesbaum.util.freenet.fcp.FcpException;
import de.todesbaum.util.freenet.fcp.FcpKeyCollisionException;
import de.todesbaum.util.freenet.fcp.FcpRuntimeException;


/**
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:UploadRequestInsertService.java 220 2006-03-06 14:46:23Z bombe $
 */
public class UploadRequestInsertService extends AbstractEngineService {

	public static final String NAME = "Upload Request Insert Service";
	private List<UploadRequestInsertListener> uploadRequestInsertListeners;
	private UploadRequest uploadRequest;

	private TimeZone gmt;
	private int lastHour;
	private boolean insertNow;

	public String getName() {
		return NAME;
	}

	public void addUploadRequestInsertListener(UploadRequestInsertListener uploadRequestInsertListener) {
		if (!uploadRequestInsertListeners.contains(uploadRequestInsertListener)) {
			uploadRequestInsertListeners.add(uploadRequestInsertListener);
		}
	}

	public void removeUploadRequestInsertListener(UploadRequestInsertListener uploadRequestInsertListener) {
		uploadRequestInsertListeners.remove(uploadRequestInsertListener);
	}

	/**
	 * Notifies listeners that the insertion of an upload request is about to
	 * start.
	 * 
	 * @param key
	 *            The key of the upload request
	 */
	private void fireUploadRequestInsertStarted(String key) {
		for (UploadRequestInsertListener uploadRequestInsertListener: uploadRequestInsertListeners) {
			uploadRequestInsertListener.uploadRequestInsertStarted(key);
		}
	}

	/**
	 * Notifies listeners that the insertion of an upload request failed.
	 * 
	 * @param key
	 *            The key of the upload request
	 * @param exception
	 *            The exception that prevented insertion
	 */
	private void fireUploadRequestInsertFailed(String key, Exception exception) {
		for (UploadRequestInsertListener uploadRequestInsertListener: uploadRequestInsertListeners) {
			uploadRequestInsertListener.uploadRequestInsertFailed(key, exception);
		}
	}

	/**
	 * Notifies listeners that the insertion of an upload request is complete.
	 * 
	 * @param key
	 *            The key of the upload request
	 */
	private void fireUploadRequestInsertComplete(String key) {
		for (UploadRequestInsertListener uploadRequestInsertListener: uploadRequestInsertListeners) {
			uploadRequestInsertListener.uploadRequestInsertComplete(key);
		}
	}
	
	public void setClient(Client localClient) {
		uploadRequest = new UploadRequest(localClient.getClientId(), System.currentTimeMillis());
	}
	
	public void addUploadRequest(String key, int blockCount) {
		BitSet allBlocks = new BitSet(blockCount);
		allBlocks.set(0, blockCount);
		addUploadRequest(key, allBlocks);
	}

	public void addUploadRequest(String key, BitSet blockList) {
		synchronized (syncObject) {
			uploadRequest.addKey(key, blockList);
			uploadRequest.touch();
			insertNow = true;
			syncObject.notify();
		}
	}
	
	public void removeUploadRequest(String key) {
		synchronized (syncObject) {
			uploadRequest.remove(key);
			uploadRequest.touch();
			insertNow = true;
			syncObject.notify();
		}
	}

	public UploadRequest getUploadRequest() {
		return uploadRequest;
	}

	public void init() {
		super.init();
		uploadRequestInsertListeners = new ArrayList<UploadRequestInsertListener>();
		gmt = TimeZone.getTimeZone("GMT-00:00");
		lastHour = -1;
		insertNow = false;
	}

	protected void work() {
		boolean inserted;
		synchronized (syncObject) {
			inserted = !insertNow;
			insertNow = false;
		}
		Calendar c = Calendar.getInstance();
		int hour = c.get(Calendar.HOUR_OF_DAY);
		inserted &= (hour == lastHour);
		
		while ((uploadRequest != null) && (uploadRequest.size() > 0) && !inserted) {
			Calendar currentDate = Calendar.getInstance(gmt);
			String keyPrefix = "KSK@" + networkName + "/uploadrequest/" + DateFormatter.getKeyPortion(currentDate) + "/";
			SimpleXML uploadRequestXML = createUploadRequestProperties(uploadRequest);
			byte[] uploadRequestBytes = XML.transformToByteArray(uploadRequestXML.getDocument());
			Data uploadRequestData = new Data(new byte[0], uploadRequestBytes);
			int index = 0;
			while (!stopping && !inserted) {
				String key = keyPrefix + Hex.toHex(index);
				fireUploadRequestInsertStarted(key);
				try {
					engine.putKey(key, hopsToLive, uploadRequestData);
					inserted = true;
					lastHour = hour;
					fireUploadRequestInsertComplete(key);
				} catch (FcpKeyCollisionException fkce1) {
					index++;
				} catch (FcpRuntimeException fre1) {
					fireUploadRequestInsertFailed(key, fre1);
				} catch (IOException ioe1) {
					fireUploadRequestInsertFailed(key, ioe1);
				} catch (FcpException fe1) {
					fireServiceAborted(fe1);
					stopping = true;
				} catch (EngineStoppedException ese1) {
					fireServiceAborted(ese1);
					stopping = true;
				}
			}
		}
	}
	
	private SimpleXML createUploadRequestProperties(UploadRequest uploadRequests) {
		SimpleXML uploadRequestXML = new SimpleXML("upload-request");
		uploadRequestXML.append("client-id", uploadRequest.getClientId());
		uploadRequestXML.append("timestamp", Hex.toHex(uploadRequest.getTimestamp()));
		SimpleXML requestNodes = uploadRequestXML.append("requests");
		for (int uploadIndex = 0, uploadCount = uploadRequest.size(); uploadIndex < uploadCount; uploadIndex++) {
			SimpleXML requestNode = requestNodes.append("request");
			requestNode.append("key", uploadRequest.getKey(uploadIndex));
			requestNode.append("blocks", Hex.toHex(uploadRequest.getBlockList(uploadIndex)));
		}
		return uploadRequestXML;
	}

}
