/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.w3c.dom.Document;

import de.todesbaum.freemule.util.XML;
import de.todesbaum.util.freenet.fcp.FcpException;

/**
 * The Upload Service is responsible for uploading files to the network as a response to Upload Requests.
 * <p>
 * Every {@link de.todesbaum.freemule.application.Client} that sent an upload request and has not yet disappeared (as determined by
 * {@link de.todesbaum.freemule.application.ClientService} has an {@link de.todesbaum.freemule.application.UploadRequest} associated with it. This
 * upload request is updated by the {@link UploadRequestDiscoveryService}. Every time an upload request is discovered that is newer than the
 * currently associated upload request it replaces it. Then all registered clients and their upload requests are iterated over and a list of files to
 * upload is assembled.
 * 
 * 
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:UploadService.java 220 2006-03-06 14:46:23Z bombe $
 */
public class UploadService extends AbstractEngineService {

	public static final String NAME = "Upload Service";

	private Set<String> uploadedKeys;
	private Set<UploadListener> uploadListeners;
	private SharedFileService sharedFileService;
	private List<UploadFile> uploadFiles;
	private Map<Client, UploadRequest> clientUploadRequests;
	private Map<String, UploadFile> keyUploadFiles;
	private Map<UploadFile, BitSet> uploadFileBitSets;
	private Random random;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String getName() {
		return NAME;
	}

	/**
	 * Adds a new {@link UploadListener} to the list of registered listeners.
	 * 
	 * @param uploadListener
	 *            The {@link UploadListener} to add
	 */
	public void addUploadListener(UploadListener uploadListener) {
		uploadListeners.add(uploadListener);
	}

	/**
	 * Removes an {@link UploadListener} from the list of registered listeners.
	 * 
	 * @param uploadListener
	 *            The {@link UploadListener} to remove
	 */
	public void removeUploadListener(UploadListener uploadListener) {
		uploadListeners.remove(uploadListener);
	}

	private void fireUploadAdded(UploadFile uploadFile) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.uploadAdded(uploadFile);
		}
	}

	private void fireUploadUpdated(UploadFile uploadFile) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.uploadUpdated(uploadFile);
		}
	}

	private void fireUploadRemoved(UploadFile uploadFile) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.uploadRemoved(uploadFile);
		}
	}

	private void fireBlockUploadStarted(UploadFile uploadFile, int blockIndex) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.blockUploadStarted(uploadFile, blockIndex);
		}
	}

	private void fireBlockUploadFailed(UploadFile uploadFile, int blockIndex, Exception exception) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.blockUploadFailed(uploadFile, blockIndex, exception);
		}
	}

	private void fireBlockUploadComplete(UploadFile uploadFile, int blockIndex) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.blockUploadComplete(uploadFile, blockIndex);
		}
	}

	private void fireManifestUploadStarted(UploadFile uploadFile) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.manifestUploadStarted(uploadFile);
		}
	}

	private void fireManifestUploadFailed(UploadFile uploadFile, Exception exception) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.manifestUploadFailed(uploadFile, exception);
		}
	}

	private void fireManifestUploadComplete(UploadFile uploadFile) {
		for (UploadListener uploadListener: uploadListeners) {
			uploadListener.manifestUploadComplete(uploadFile);
		}
	}

	public void setSharedFileService(SharedFileService sharedFileService) {
		this.sharedFileService = sharedFileService;
	}

	public void addUploadRequest(Client remoteClient, UploadRequest uploadRequest) {
		UploadRequest clientUploadRequest = clientUploadRequests.get(remoteClient);
		if ((clientUploadRequest == null) || (uploadRequest.getTimestamp() > clientUploadRequest.getTimestamp())) {
			clientUploadRequests.put(remoteClient, uploadRequest);
			updateUploads();
		}
	}

	public void removeClientUploadRequest(Client remoteClient) {
		if (clientUploadRequests.containsValue(remoteClient)) {
			clientUploadRequests.remove(remoteClient);
			updateUploads();
		}
	}

	private synchronized void updateUploads() {
		List<UploadFile> tempUploadFiles = new ArrayList<UploadFile>();
		Map<String, UploadFile> tempKeyUploadFiles = new HashMap<String, UploadFile>();
		Map<UploadFile, BitSet> tempUploadFileBitSets = new HashMap<UploadFile, BitSet>();
		Iterator<Client> clients = clientUploadRequests.keySet().iterator();
		while (clients.hasNext()) {
			Client remoteClient = clients.next();
			UploadRequest clientUploadRequest = clientUploadRequests.get(remoteClient);
			for (int keyIndex = 0, keyCount = clientUploadRequest.size(); keyIndex < keyCount; keyIndex++) {
				String uploadKey = clientUploadRequest.getKey(keyIndex);
				if (uploadedKeys.contains(uploadKey)) {
					continue;
				}
				SharedFile sharedFile = sharedFileService.getSharedFile(uploadKey);
				if (sharedFile == null) {
					continue;
				}
				UploadFile uploadFile = new UploadFile(sharedFile);
				tempKeyUploadFiles.put(uploadKey, uploadFile);
				tempUploadFiles.add(uploadFile);
				BitSet existingUploadFileBitSet = tempUploadFileBitSets.get(uploadFile);
				if (existingUploadFileBitSet != null) {
					existingUploadFileBitSet.or(clientUploadRequest.getBlockList(keyIndex));
				} else {
					tempUploadFileBitSets.put(uploadFile, clientUploadRequest.getBlockList(keyIndex));
				}
			}
		}
		// check for new upload files
		Iterator<UploadFile> tempUploadFilesIterator = tempUploadFiles.iterator();
		while (tempUploadFilesIterator.hasNext()) {
			UploadFile newUploadFile = tempUploadFilesIterator.next();
			if (!uploadFiles.contains(newUploadFile)) {
				keyUploadFiles.put(newUploadFile.getKey(), newUploadFile);
				uploadFileBitSets.put(newUploadFile, tempUploadFileBitSets.get(newUploadFile));
				uploadFiles.add(newUploadFile);
				newUploadFile.setStatus(FileStatus.activated);
				newUploadFile.setManifestStatus(BlockStatus.notAvailable);
				for (int blockIndex = 0, blockCount = newUploadFile.getBlockCount(); blockIndex < blockCount; blockIndex++) {
					newUploadFile.setBlockStatus(blockIndex, BlockStatus.notAvailable);
				}
				fireUploadAdded(newUploadFile);
			} else {
				BitSet bitSet = uploadFileBitSets.get(newUploadFile);
				if (!bitSet.equals(tempUploadFileBitSets.get(newUploadFile))) {
					bitSet.clear();
					bitSet.or(tempUploadFileBitSets.get(newUploadFile));
					fireUploadUpdated(newUploadFile);
				}
			}
		}
		// now check for delete upload files
		Iterator<UploadFile> uploadFilesIterator = uploadFiles.iterator();
		List<UploadFile> uploadFilesToRemove = new ArrayList<UploadFile>();
		while (uploadFilesIterator.hasNext()) {
			UploadFile uploadFile = uploadFilesIterator.next();
			if (!tempUploadFiles.contains(uploadFile)) {
				uploadFilesToRemove.add(uploadFile);
			}
		}
		for (UploadFile uploadFileToRemove: uploadFilesToRemove) {
			uploadFiles.remove(uploadFileToRemove);
			keyUploadFiles.remove(uploadFileToRemove.getKey());
			uploadFileBitSets.remove(uploadFileToRemove);
			fireUploadRemoved(uploadFileToRemove);
		}
	}

	public UploadFile getUploadFile(String key) {
		for (UploadFile uploadFile: uploadFiles) {
			if (uploadFile.getKey().equals(key)) {
				return uploadFile;
			}
		}
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init() {
		super.init();
		uploadedKeys = new HashSet<String>();
		uploadListeners = new HashSet<UploadListener>();
		uploadFiles = new ArrayList<UploadFile>();
		clientUploadRequests = new HashMap<Client, UploadRequest>();
		keyUploadFiles = new HashMap<String, UploadFile>();
		uploadFileBitSets = new HashMap<UploadFile, BitSet>();
		random = new Random();
		setWaitPeriod(1000);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void destroy() {
		random = null;
		uploadedKeys.clear();
		uploadedKeys = null;
		uploadFileBitSets.clear();
		uploadFileBitSets = null;
		keyUploadFiles.clear();
		keyUploadFiles = null;
		clientUploadRequests.clear();
		clientUploadRequests = null;
		uploadListeners.clear();
		uploadListeners = null;
		uploadFiles.clear();
		uploadFiles = null;
		super.destroy();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void work() {
		List<UploadFile> tempUploadFiles = null;
		synchronized (this) {
			tempUploadFiles = Collections.unmodifiableList(uploadFiles);
		}
		if (tempUploadFiles.size() > 0) {
			UploadInformation nextUploadInformation = findNextUpload(tempUploadFiles);
			if (nextUploadInformation != null) {
				UploadFile uploadFile = nextUploadInformation.getUploadFile();
				int blockIndex = -1;
				try {
					if (nextUploadInformation.isManifest()) {
						Document manifest = engine.createManifest(uploadFile);
						byte[] manifestBytes = XML.transformToByteArray(manifest);
						fireManifestUploadStarted(uploadFile);
						uploadFile.setManifestStatus(BlockStatus.requesting);
						engine.putKey("CHK@", hopsToLive, true, new byte[0], manifestBytes, true);
						uploadFile.setManifestStatus(BlockStatus.available);
						fireManifestUploadComplete(uploadFile);
					} else {
						blockIndex = nextUploadInformation.getBlockNumber();
						fireBlockUploadStarted(uploadFile, blockIndex);
						uploadFile.setBlockStatus(blockIndex, BlockStatus.requesting);
						engine.putKey("CHK@", hopsToLive, true, uploadFile.getBlockInputStream(blockIndex), Math.min(FreenetFile.BLOCKSIZE, uploadFile.getLength() - blockIndex * FreenetFile.BLOCKSIZE), new byte[0], true);
						uploadFile.setBlockStatus(blockIndex, BlockStatus.available);
						BitSet uploadFileBitSet = uploadFileBitSets.get(uploadFile);
						uploadFileBitSet.clear(blockIndex);
						fireBlockUploadComplete(uploadFile, blockIndex);
						if (uploadFileBitSet.cardinality() == 0) {
							uploadedKeys.add(uploadFile.getKey());
							uploadFile.setStatus(FileStatus.complete);
							fireUploadUpdated(uploadFile);
						}
					}
				} catch (FcpException fcpe1) {
					if (blockIndex > -1) {
						fireBlockUploadFailed(uploadFile, blockIndex, fcpe1);
					} else {
						fireManifestUploadFailed(uploadFile, fcpe1);
					}
				} catch (IOException ioe1) {
					if (blockIndex > -1) {
						fireBlockUploadFailed(uploadFile, blockIndex, ioe1);
					} else {
						fireManifestUploadFailed(uploadFile, ioe1);
					}
				} catch (EngineStoppedException ese1) {
					fireServiceAborted(ese1);
					stopping = true;
				}

			}
		}
	}

	private UploadInformation findNextUpload(List<UploadFile> tempUploadFiles) {
		List<UploadFile> availableUploadFiles = getAvailableUploadFiles(tempUploadFiles);
		if (availableUploadFiles.size() > 0) {
			int fileIndex = random.nextInt(availableUploadFiles.size());
			UploadFile uploadFile = availableUploadFiles.get(fileIndex);
			if (uploadFile.getManifestStatus() == BlockStatus.notAvailable) {
				return new UploadInformation(uploadFile);
			}
			List<Integer> availableBlocks = getAvailableBlocks(uploadFile);
			if (availableBlocks.size() > 0) {
				int blockIndex = random.nextInt(availableBlocks.size());
				int blockNumber = availableBlocks.get(blockIndex);
				return new UploadInformation(uploadFile, blockNumber);
			}
		}
		return null;
	}

	/**
	 * Returns the files that are currently uploaded.
	 * 
	 * @return The files that are currently uploaded
	 */
	public UploadFile[] getUploadFiles() {
		return uploadFiles.toArray(new UploadFile[uploadFiles.size()]);
	}

	private List<UploadFile> getAvailableUploadFiles(List<UploadFile> uploadFiles) {
		List<UploadFile> availableUploadFiles = new ArrayList<UploadFile>();
		for (UploadFile uploadFile: uploadFiles) {
			if (uploadFile.getStatus() != FileStatus.activated) {
				continue;
			}
			if (uploadFile.getManifestStatus() == BlockStatus.notAvailable) {
				availableUploadFiles.add(uploadFile);
				continue;
			}
			BitSet uploadFileBitSet = uploadFileBitSets.get(uploadFile);
			if (uploadFileBitSet.cardinality() > 0) {
				availableUploadFiles.add(uploadFile);
			}
		}
		return availableUploadFiles;
	}

	private List<Integer> getAvailableBlocks(UploadFile uploadFile) {
		BitSet uploadFileBitSet = uploadFileBitSets.get(uploadFile);
		List<Integer> blockList = new ArrayList<Integer>();
		for (int bitIndex = uploadFileBitSet.nextSetBit(0); bitIndex >= 0; bitIndex = uploadFileBitSet.nextSetBit(bitIndex + 1)) {
			blockList.add(bitIndex);
		}
		return blockList;
	}

}
