/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.gui;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableModel;

import de.todesbaum.freemule.application.Client;
import de.todesbaum.freemule.application.DownloadService;
import de.todesbaum.freemule.gui.SearchResultTableModel.ResultFile;


/**
 * A {@link JPanel panel} that lets the user search for files.
 * 
 * @author <a href="mailto:droden@gmail.com">David Roden</a>
 * @version $Id:SearchResultPanel.java 221 2006-03-06 14:46:49Z bombe $
 */
public class SearchResultPanel extends JPanel implements ListSelectionListener, MouseListener {

	private JFrame parent;
	private JTable resultTable;
	private SearchResultTableModel resultTableModel;
	private TableSorter resultTableSorter;
	private Action downloadAction;
	private Action showCommentsAction;
	private Action showFilenamesAction;
	private JPopupMenu popupMenu;
	
	private DownloadService downloadService;

	public SearchResultPanel(JFrame parent) {
		super(new BorderLayout(5, 5));
		this.parent = parent;
		initializeActions();
		initializeComponents();
		initializeLayout();
	}

	public void setDownloadService(DownloadService downloadService) {
		this.downloadService = downloadService;
	}

	private void initializeActions() {
		downloadAction = new AbstractAction("Download") {

			public void actionPerformed(ActionEvent actionEvent) {
				actionDownload();
			}
		};
		downloadAction.setEnabled(false);
		showCommentsAction = new AbstractAction("Show Comments") {

			public void actionPerformed(ActionEvent actionEvent) {
				actionShowComments();
			}
		};
		showCommentsAction.setEnabled(false);
		showFilenamesAction = new AbstractAction("Show Filenames") {

			public void actionPerformed(ActionEvent actionEvent) {
				actionShowFilenames();
			}
		};
		showFilenamesAction.setEnabled(false);
	}

	private void initializeComponents() {
		resultTableModel = new SearchResultTableModel();
		resultTableSorter = new TableSorter(resultTableModel);
		resultTable = new JTable(resultTableSorter);
		resultTable.getSelectionModel().addListSelectionListener(this);
		resultTable.addMouseListener(this);
		resultTableSorter.setTableHeader(resultTable.getTableHeader());
		popupMenu = new JPopupMenu("Search Result");
		popupMenu.add(downloadAction);
		popupMenu.add(showCommentsAction);
		popupMenu.add(showFilenamesAction);
	}

	private void initializeLayout() {
		setBorder(new EmptyBorder(5, 5, 5, 5));
		add(new JScrollPane(resultTable), BorderLayout.CENTER);
		add(createButtonPanel(), BorderLayout.PAGE_END);
	}

	private JPanel createButtonPanel() {
		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.TRAILING, 5, 5));
		buttonPanel.setBorder(new EmptyBorder(-5, -5, -5, -5));
		buttonPanel.add(new JButton(downloadAction));
		buttonPanel.add(new JButton(showCommentsAction));
		buttonPanel.add(new JButton(showFilenamesAction));
		return buttonPanel;
	}

	public JTable getResultTable() {
		return resultTable;
	}

	public TableModel getTableModel() {
		return resultTableSorter;
	}

	public void addResultFile(ResultFile resultFile) {
		resultTableModel.addFile(resultFile);
	}

	//
	// ACTIONS
	//

	protected void actionDownload() {
		int[] selectedRows = resultTable.getSelectedRows();
		for (int selectedRow: selectedRows) {
			ResultFile resultFile = (ResultFile) resultTableSorter.getValueAt(selectedRow, -1);
			downloadService.addDownload(resultFile.getKey(), resultFile.getName(), resultFile.getLength());
		}
	}

	protected void actionShowComments() {
		int selectedRow = resultTable.getSelectedRow();
		ResultFile resultFile = (ResultFile) resultTableSorter.getValueAt(selectedRow, -1);
		String[] comments = resultTableModel.getComments(resultFile.getKey());
		int[] ratings = resultTableModel.getRatings(resultFile.getKey());
		Client[] clients = resultTableModel.getCommentClients(resultFile.getKey());
		SearchResultCommentDialog dialog = new SearchResultCommentDialog(parent, resultFile, clients, comments, ratings);
		dialog.setVisible(true);
	}

	protected void actionShowFilenames() {
		int selectedRow = resultTable.getSelectedRow();
		ResultFile resultFile = (ResultFile) resultTableSorter.getValueAt(selectedRow, -1);
		String[] filenames = resultTableModel.getFilenames(resultFile.getKey());
		SearchResultFilenameDialog dialog = new SearchResultFilenameDialog(parent, resultFile, filenames);
		dialog.setVisible(true);
	}

	/**
	 * {@inheritDoc}
	 */
	public void valueChanged(ListSelectionEvent e) {
		int selectedRowCount = resultTable.getSelectedRowCount();
		downloadAction.setEnabled(selectedRowCount > 0);
		if (selectedRowCount == 1) {
			ResultFile resultFile = (ResultFile) resultTableSorter.getValueAt(resultTable.getSelectedRow(), -1);
			showCommentsAction.setEnabled(resultTableModel.hasComments(resultFile.getKey()));
			showFilenamesAction.setEnabled(resultFile.getAvailability() > 1);
		} else {
			showCommentsAction.setEnabled(false);
			showFilenamesAction.setEnabled(false);
		}
	}

	private void showPopupMenu(int x, int y) {
		popupMenu.show(resultTable, x, y);
	}

	/**
	 * {@inheritDoc}
	 */
	public void mouseClicked(MouseEvent e) {
		if (e.isPopupTrigger()) {
			showPopupMenu(e.getX(), e.getY());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void mousePressed(MouseEvent e) {
		if (e.isPopupTrigger()) {
			showPopupMenu(e.getX(), e.getY());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void mouseReleased(MouseEvent e) {
	}

	/**
	 * {@inheritDoc}
	 */
	public void mouseEntered(MouseEvent e) {
	}

	/**
	 * {@inheritDoc}
	 */
	public void mouseExited(MouseEvent e) {
	}

}
