/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.main;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Manages {@link de.todesbaum.freemule.main.LoggingPreferences logging preferences} for a {@link de.todesbaum.freemule.application.Service}. Logging can set to logging to
 * file, logging to the GUI, or to both for every event of a service. Events are split up into different groups to ease configuration from a GUI.
 * 
 * @author David Roden &lt;droden@gmail.com&gt;
 * @version $Id:LoggingOptions.java 221 2006-03-06 14:46:49Z bombe $
 */
public class LoggingOptions {

	/** The name of the service. */
	private String serviceName;

	/** A list of all event group names. */
	private List<String> groupNames;

	/** Maps event names to group names. */
	private Map<String, String[]> groupEventNames;

	/** Maps event groups to event names. */
	private Map<String, String> eventNameGroups;

	/** Maps event names to GUI logging states. */
	private Map<String, Boolean> logToGuiEvents;

	/** Maps event names to file logging states. */
	private Map<String, Boolean> logToFileEvents;

	/**
	 * Constructs new logging options for a service with the specified name.
	 * 
	 * @param serviceName
	 *            The name of the service these logging options are meant for
	 */
	public LoggingOptions(String serviceName) {
		this.serviceName = serviceName;
		this.groupNames = new ArrayList<String>();
		this.groupEventNames = new HashMap<String, String[]>();
		this.eventNameGroups = new HashMap<String, String>();
		this.logToGuiEvents = new HashMap<String, Boolean>();
		this.logToFileEvents = new HashMap<String, Boolean>();
	}

	/**
	 * Adds a new group of logging events to these options.
	 * 
	 * @param groupName
	 *            The name of the group to add
	 * @param eventNames
	 *            All event names of the group
	 */
	public void addLoggingGroup(String groupName, String[] eventNames) {
		groupNames.add(groupName);
		groupEventNames.put(groupName, eventNames);
		for (String eventName: eventNames) {
			eventNameGroups.put(eventName, groupName);
		}
	}

	/**
	 * Returns the name of the service these logging options are meant for.
	 * 
	 * @return The name of the service
	 */
	public String getServiceName() {
		return serviceName;
	}

	/**
	 * Returns all group names of these options.
	 * 
	 * @return All group names of these options
	 */
	public String[] getGroups() {
		return groupNames.toArray(new String[groupNames.size()]);
	}

	/**
	 * Returns all event names within the specified group.
	 * 
	 * @param groupName
	 *            The group to get event names for
	 * @return All event names within the specified group
	 */
	public String[] getGroupEventNames(String groupName) {
		return groupEventNames.get(groupName);
	}

	/**
	 * Sets the logging options for an event.
	 * 
	 * @param eventName
	 *            The name of the event
	 * @param logToGui
	 *            Whether to log to the GUI
	 * @param logToFile
	 *            Whether to log to a file
	 * @see #setGUILogging(String, boolean)
	 * @see #setFileLogging(String, boolean)
	 */
	public void setLogging(String eventName, boolean logToGui, boolean logToFile) {
		setGUILogging(eventName, logToGui);
		setFileLogging(eventName, logToFile);
	}

	/**
	 * Sets whether to log an event to the GUI.
	 * 
	 * @param eventName
	 *            The name of the event
	 * @param logToGui
	 *            Whether to log to the GUI
	 */
	public void setGUILogging(String eventName, boolean logToGui) {
		logToGuiEvents.put(eventName, logToGui);

	}

	/**
	 * Sets whether to log an event to a file.
	 * 
	 * @param eventName
	 *            The name of the event
	 * @param logToFile
	 *            Whether to log to a file
	 */
	public void setFileLogging(String eventName, boolean logToFile) {
		logToFileEvents.put(eventName, logToFile);

	}

	/**
	 * Returns whether to log an event to the GUI.
	 * 
	 * @param eventName
	 *            The name of the event
	 * @return Whether to log this event to the GUI
	 */
	public boolean getGUILogging(String eventName) {
		return logToGuiEvents.get(eventName);
	}

	/**
	 * Returns whether to log an event to a file.
	 * 
	 * @param eventName
	 *            The name of the event
	 * @return Whether to log this event to a file
	 */
	public boolean getFileLogging(String eventName) {
		return logToFileEvents.get(eventName);
	}

}