/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.freemule.main;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * The logging preferences for the application. Contains a list of service names and their {@link de.todesbaum.freemule.main.LoggingOptions} as well as the name of
 * the log file.
 * 
 * @author David Roden &lt;droden@gmail.com&gt;
 * @version $Id:LoggingPreferences.java 221 2006-03-06 14:46:49Z bombe $
 */
public class LoggingPreferences {

	/** The name of the log file. */
	private String logfile;

	/** The names of the services. */
	private List<String> services;

	/** The {@link de.todesbaum.freemule.main.LoggingOptions} for a service. */
	private Map<String, LoggingOptions> serviceLoggingOptions;

	/**
	 * Creates new logging preferences.
	 */
	public LoggingPreferences() {
		this.services = new ArrayList<String>();
		this.serviceLoggingOptions = new HashMap<String, LoggingOptions>();
	}

	/**
	 * Sets the name of the log file to the specified name.
	 * 
	 * @param logfile
	 *            The name of the log file
	 */
	public void setLogfile(String logfile) {
		this.logfile = logfile;
	}

	/**
	 * Returns the name of the log file.
	 * 
	 * @return The name of the log file
	 */
	public String getLogfile() {
		return logfile;
	}

	/**
	 * Returns all service names.
	 * 
	 * @return All service names
	 */
	public String[] getServiceNames() {
		return services.toArray(new String[services.size()]);
	}

	/**
	 * Returns the logging options for the specified service.
	 * 
	 * @param serviceName
	 *            The name of the service
	 * @return The logging options for the specified service
	 */
	public LoggingOptions getServiceLoggingOptions(String serviceName) {
		return serviceLoggingOptions.get(serviceName);
	}

	/**
	 * Adds logging options for a service.
	 * 
	 * @param loggingOptions
	 *            The logging options to add
	 * @see LoggingOptions#getServiceName()
	 */
	public void addServiceLoggingOptions(LoggingOptions loggingOptions) {
		services.add(loggingOptions.getServiceName());
		serviceLoggingOptions.put(loggingOptions.getServiceName(), loggingOptions);
	}

	/**
	 * Returns whether to log an event of a service to the GUI.
	 * 
	 * @param serviceName
	 *            The name of the service
	 * @param eventName
	 *            The name of the event
	 * @return Whether to log this event to the GUI
	 * @see LoggingOptions#getGUILogging(String)
	 */
	public boolean getGUILogging(String serviceName, String eventName) {
		return getServiceLoggingOptions(serviceName).getGUILogging(eventName);
	}

	/**
	 * Returns whether to log an event of a service to a file.
	 * 
	 * @param serviceName
	 *            The name of the service
	 * @param eventName
	 *            The name of the event
	 * @return Whether to log this event to a file
	 * @see LoggingOptions#getFileLogging(String)
	 */
	public boolean getFileLogging(String serviceName, String eventName) {
		return getServiceLoggingOptions(serviceName).getFileLogging(eventName);
	}

}