/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.util.freenet.fcp;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * FCP <code>ClientPut</code> implementation command.
 * 
 * @see de.todesbaum.util.freenet.fcp.FcpClientGet
 * @author David Roden <droden@gmail.com>
 * @version $Id: FcpClientPut.java 218 2006-03-06 14:43:35Z  $
 */
public class FcpClientPut extends FcpCommand {

	/**
	 * The URI under which to store the data.
	 */
	private final String uri;

	/**
	 * The hops-to-live value used for the insertion.
	 */
	private final int hopsToLive;

	/**
	 * Whether to skip the local data store on insertion.
	 */
	private final boolean removeLocalKey;

	/**
	 * The {@link InputStream} to read the data from.
	 */
	private InputStream dataInputStream;

	/**
	 * The length of the data input stream.
	 */
	private long dataLength;

	/**
	 * The {@link InputStream} to read the metadata from.
	 */
	private InputStream metadataInputStream;

	/**
	 * The length of the metadata input stream.
	 */
	private long metadataLength;

	/**
	 * The resulting URI.
	 */
	private String finalURI;

	/**
	 * The private key of the resulting URI.
	 */
	private String privateKey;

	/**
	 * The public key of the resulting URI.
	 */
	private String publicKey;

	/**
	 * Creates a new FCP <code>ClientPut</code> command with the specified
	 * parameters. If the URI of the data is not known before insertion use
	 * <code>"CHK@"</code> as URI.
	 * 
	 * @param uri
	 *            The URI under which to save the data
	 * @param hopsToLive
	 *            The hops-to-live value of the insert
	 * @param data
	 *            The data to insert
	 */
	public FcpClientPut(String uri, int hopsToLive, byte[] data) {
		this(uri, hopsToLive, false, data);
	}

	/**
	 * Creates a new FCP <code>ClientPut</code> command with the specified
	 * parameters. If the URI of the data is not known before insertion use
	 * <code>"CHK@"</code> as URI.
	 * 
	 * @param uri
	 *            The URI under which to save the data
	 * @param hopsToLive
	 *            The hops-to-live value of the insert
	 * @param removeLocalKey
	 *            <code>true</code> if this insert is to skip the local data
	 *            store, <code>false</code> otherwise
	 * @param data
	 *            The data to insert
	 */
	public FcpClientPut(String uri, int hopsToLive, boolean removeLocalKey, byte[] data) {
		this(uri, hopsToLive, removeLocalKey, data, new byte[0]);
	}

	/**
	 * Creates a new FCP <code>ClientPut</code> command with the specified
	 * parameters. If the URI of the data is not known before insertion use
	 * <code>"CHK@"</code> as URI.
	 * 
	 * @param uri
	 *            The URI under which to save the data
	 * @param hopsToLive
	 *            The hops-to-live value of the insert
	 * @param data
	 *            The data to insert
	 * @param metadata
	 *            The metadata to insert
	 */
	public FcpClientPut(String uri, int hopsToLive, byte[] data, byte[] metadata) {
		this(uri, hopsToLive, false, data, metadata);
	}

	/**
	 * Creates a new FCP <code>ClientPut</code> command with the specified
	 * parameters. If the URI of the data is not known before insertion use
	 * <code>"CHK@"</code> as URI.
	 * 
	 * @param uri
	 *            The URI under which to save the data
	 * @param hopsToLive
	 *            The hops-to-live value of the insert
	 * @param removeLocalKey
	 *            <code>true</code> if this insert is to skip the local data
	 *            store, <code>false</code> otherwise
	 * @param data
	 *            The data to insert
	 * @param metadata
	 *            The metadata to insert
	 */
	public FcpClientPut(String uri, int hopsToLive, boolean removeLocalKey, byte[] data, byte[] metadata) {
		this(uri, hopsToLive, removeLocalKey, new ByteArrayInputStream(data), data.length, new ByteArrayInputStream(metadata), metadata.length);
	}

	/**
	 * Creates a new FCP <code>ClientPut</code> command with the specified
	 * parameters. If the URI of the data is not known before insertion use
	 * <code>"CHK@"</code> as URI.
	 * 
	 * @param uri
	 *            The URI under which to save the data
	 * @param hopsToLive
	 *            The hops-to-live value of the insert
	 * @param removeLocalKey
	 *            <code>true</code> if this insert is to skip the local data
	 *            store, <code>false</code> otherwise
	 * @param dataInputStream
	 *            The {@link InputStream} to read the data from
	 * @param dataLength
	 *            The length of the data input stream
	 * @param metadata
	 *            The metadata to insert
	 */
	public FcpClientPut(String uri, int hopsToLive, boolean removeLocalKey, InputStream dataInputStream, long dataLength, byte[] metadata) {
		this(uri, hopsToLive, removeLocalKey, dataInputStream, dataLength, new ByteArrayInputStream(metadata), metadata.length);
	}

	/**
	 * Creates a new FCP <code>ClientPut</code> command with the specified
	 * parameters. If the URI of the data is not known before insertion use
	 * <code>"CHK@"</code> as URI.
	 * 
	 * @param uri
	 *            The URI under which to save the data
	 * @param hopsToLive
	 *            The hops-to-live value of the insert
	 * @param removeLocalKey
	 *            <code>true</code> if this insert is to skip the local data
	 *            store, <code>false</code> otherwise
	 * @param dataInputStream
	 *            The {@link InputStream} to read the data from
	 * @param dataLength
	 *            The length of the data input stream
	 * @param metadataInputStream
	 *            The {@link InputStream} to read the metadata from
	 * @param metadataLength
	 *            The length of the metadata input stream
	 */
	public FcpClientPut(String uri, int hopsToLive, boolean removeLocalKey, InputStream dataInputStream, long dataLength, InputStream metadataInputStream, long metadataLength) {
		this.uri = uri;
		this.hopsToLive = hopsToLive;
		this.removeLocalKey = removeLocalKey;
		this.dataInputStream = dataInputStream;
		this.dataLength = dataLength;
		this.metadataInputStream = metadataInputStream;
		this.metadataLength = metadataLength;
	}

	/**
	 * {@inheritDoc}
	 */
	public String getName() {
		return "ClientPut";
	}

	/**
	 * Returns the resulting URI. Only useful if the URI was <code>"CHK@"</code>
	 * before inserting.
	 * 
	 * @return The resulting URI
	 */
	public String getFinalURI() {
		return finalURI;
	}

	/**
	 * Returns the private key of the URI.
	 * 
	 * @return The private key of the URI
	 */
	public String getPrivateKey() {
		return privateKey;
	}

	/**
	 * Returns the public key of the URI.
	 * 
	 * @return The public key of the URI
	 */
	public String getPublicKey() {
		return publicKey;
	}

	/**
	 * {@inheritDoc}
	 */
	protected boolean writeCommand(OutputStream outputStream) throws IOException {
		outputStream.write(("URI=" + uri + "\r\n").getBytes());
		outputStream.write(("HopsToLive=" + hopsToLive + "\r\n").getBytes());
		outputStream.write(("RemoveLocalKey=" + removeLocalKey + "\r\n").getBytes());
		outputStream.write(("DataLength=" + encodeHex(dataLength + metadataLength) + "\r\n").getBytes());
		if (metadataLength > 0) {
			outputStream.write(("MetadataLength=" + encodeHex(metadataLength) + "\r\n").getBytes());
		}
		outputStream.write("Data\r\n".getBytes());
		copyStream(outputStream, metadataInputStream, metadataLength);
		copyStream(outputStream, dataInputStream, dataLength);
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	protected void handleFcpMessage(FcpMessage fcpMessage) throws FcpException {
		String fcpMessageName = fcpMessage.getName();
		if (fcpMessageName.equals("URIError")) {
			throw new FcpURIErrorException(fcpMessage, fcpMessage.getParameter("Reason"));
		} else if (fcpMessageName.equals("Restarted") || fcpMessageName.equals("Pending")) {
			fireFcpEvent(fcpMessageName + ": " + fcpMessage.getParameter("Reason"));
		} else if (fcpMessageName.equals("RouteNotFound")) {
			throw new FcpRouteNotFoundException(fcpMessage);
		} else if (fcpMessageName.equals("KeyCollision")) {
			complete = true;
			throw new FcpKeyCollisionException(fcpMessage);
		} else if (fcpMessageName.equals("SizeError")) {
			throw new FcpSizeErrorException(fcpMessage);
		} else if (fcpMessageName.equals("Success")) {
			complete = true;
			finalURI = fcpMessage.getParameter("URI");
			privateKey = fcpMessage.getParameter("PrivateKey");
			publicKey = fcpMessage.getParameter("PublicKey");
		}
	}

}
