/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.util.freenet.fcp;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * The base class for all FCP commands.
 * 
 * @author David Roden <droden@gmail.com>
 * @version $Id: FcpCommand.java 218 2006-03-06 14:43:35Z  $
 */
public abstract class FcpCommand {

	/**
	 * Whether the processing of this command is complete.
	 */
	protected boolean complete = false;

	/**
	 * A {@link List} containing all event listeners.
	 */
	private List/* <FcpEventListener> */fcpEventListeners = new ArrayList/* <FcpEventListener> */();

	/**
	 * Returns the name of this command. <strong>Note:</strong> This name is
	 * also sent to the Freenet node as command so this name can not be
	 * arbitrary!
	 * 
	 * @return The name of this command
	 */
	public abstract String getName();

	/**
	 * Handles all command-specified output. The implementing class can assume
	 * that the physical and logical connection setup has already been performed
	 * and that the name of the command (including <code>CRLF</code>) has
	 * already been sent.
	 * 
	 * @param outputStream
	 *            The stream to output command-specific options and data
	 * @return <code>true</code> if the {@link FcpClient FCP client} has to
	 *         sent the <code>"EndMessage"</code> after this command's output,
	 *         <code>false</code> if no further output is necessary
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	protected abstract boolean writeCommand(OutputStream outputStream) throws IOException;

	/**
	 * If the {@link FcpClient FCP client} receives a new message from the
	 * Freenet node it's parsed into an {@link FcpMessage} object and handed
	 * over to the command. The command has to extract necessary information and
	 * sets {@link #complete} to <code>true</code> if the processing is done.
	 * <strong>Note:</strong> The command has to save it's internal state to
	 * handle the possibly multiple calls to this method correctly!
	 * 
	 * @param fcpMessage
	 *            The message received from the node
	 * @throws FcpException
	 *             if an FCP protocol error occurs
	 */
	protected abstract void handleFcpMessage(FcpMessage fcpMessage) throws FcpException, IOException;

	/**
	 * Returns whether this command completed successfully.
	 * 
	 * @return <code>true</code> if this command completed successfully,
	 *         <code>false</code> otherwise
	 */
	public final boolean isComplete() {
		return complete;
	}

	/**
	 * Add an {@link FcpEventListener} to this command's list of listeners.
	 * 
	 * @param fcpEventListener
	 *            The listener to add
	 */
	public final void addFcpEventListener(FcpEventListener fcpEventListener) {
		if (!fcpEventListeners.contains(fcpEventListener)) {
			fcpEventListeners.add(fcpEventListener);
		}
	}

	/**
	 * Removes an {@link FcpEventListener} to this command's list of listeners.
	 * 
	 * @param fcpEventListener
	 *            The listener to remove
	 */
	public final void removeFcpEventListener(FcpEventListener fcpEventListener) {
		fcpEventListeners.remove(fcpEventListener);
	}

	/**
	 * Parses a {@link String} containing a hexadecimal (base 16) number to a
	 * <code>long</code>. This is a convenience method for the internal
	 * handling of the received FCP messages as the FC protocol is heavily based
	 * on hexadecimal numbers.
	 * 
	 * @param number
	 *            The string to parse
	 * @return The parsed number, or <code>0</code> if <code>number</code>
	 *         is <code>null</code>
	 * @throws NumberFormatException
	 *             if <code>number</code> does not contain a valid hexadecimal
	 *             number
	 */
	public static final long parseHex(String number) {
		if (number == null) {
			return 0;
		}
		return Long.parseLong(number, 16);
	}

	/**
	 * Encodes a <code>long</code> to a hexadecimal number. This is a
	 * convenience method for the internal creation of FCP messages as the FC
	 * protocol is heavily based on hexadecimal numbers.
	 * 
	 * @param number
	 *            The number to encode
	 * @return The encoded number
	 */
	public static final String encodeHex(long number) {
		return Long.toHexString(number);
	}

	/**
	 * Encodes a {@link List} of {@link Integer}s to a comma-separated list of
	 * hexadecimal numbers. This is a convenience method for the internal
	 * creation of FCP messages as the FC protocol is heavily based on
	 * hexadecimal numbers.
	 * 
	 * @param blockList
	 *            The list of numbers to encode
	 * @return The encoded list
	 */
	public static final String encodeList(List/* <Integer> */blockList) {
		StringBuffer resultBuffer = new StringBuffer();
		for (int index = 0, size = blockList.size(); index < size; index++) {
			if (resultBuffer.length() > 0) {
				resultBuffer.append(',');
			}
			resultBuffer.append(encodeHex(((Integer) blockList.get(index)).intValue()));
		}
		return resultBuffer.toString();
	}

	/**
	 * If the command receives an FCP Restarting or Pending message this message
	 * will forward it to all registered listeners.
	 * 
	 * @param message
	 *            The message to forward
	 */
	protected final void fireFcpEvent(String message) {
		Iterator listeners = fcpEventListeners.iterator();
		while (listeners.hasNext()) {
			FcpEventListener fcpEventListener = (FcpEventListener) listeners.next();
			fcpEventListener.fcpGotMessage(this, message);
		}
	}
	
	protected void copyStream(OutputStream outputStream, InputStream inputStream, long length) throws IOException {
		long remaining = length;
		while (remaining > 0) {
			byte[] buffer = new byte[256 * 1024];
			int read = inputStream.read(buffer, 0, (int) Math.min(256 * 1024, remaining));
			if (read != -1) {
				outputStream.write(buffer, 0, read);
				remaining -= read;
			} else {
				throw new EOFException("stream ended prematurely.");
			}
		}
	}

}
