/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.util.freenet.fcp;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * @author David Roden <droden@gmail.com>
 * @version $Id: FcpMessage.java 218 2006-03-06 14:43:35Z  $
 */
public class FcpMessage {

	private String name = null;
	private List/*<String>*/ parameters = new ArrayList/*<String>*/();
	private Map/*<String, String>*/ values = new HashMap/*<String, String>*/();
	private byte[] data = null;

	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}

	/**
	 * @param name
	 *                  The name to set.
	 */
	public void setName(String name) {
		this.name = name;
	}

	public void addParameter(String key, String value) {
		parameters.add(key);
		values.put(key, value);
	}

	public void removeParameter(String key) {
		parameters.remove(key);
		values.remove(key);
	}

	public String getParameter(String key) {
		return (String) values.get(key);
	}
	
	public void setData(byte[] data) {
		this.data = data;
	}

	public byte[] getData() {
		if (data == null) {
			return null;
		}
		byte[] result = new byte[data.length];
		System.arraycopy(data, 0, result, 0, data.length);
		return result;
	}

	public void writeMessage(OutputStream outputStream) throws IOException {
		writeLine(outputStream, name);
		Iterator keys = parameters.iterator();
		while (keys.hasNext()) {
			Object key = keys.next();
			String value = (String) values.get(key);
			writeLine(outputStream, key + "=" + value);
		}
		writeLine(outputStream, "EndMessage");
	}

	private void writeLine(OutputStream outputStream, String line) throws IOException {
		outputStream.write((line + "\r\n").getBytes());
	}

	public static FcpMessage readMessage(InputStream inputStream) throws IOException {
		FcpMessage fcpMessage = new FcpMessage();
		boolean finished = false;
		do {
			String line = readLine(inputStream);
			if (line == null) {
				throw new IOException("EOF");
			}
			if (line.equals("EndMessage")) {
				finished = true;
			} else if (line.equals("Data")) {
				int length;
				if (fcpMessage.getParameter("Length") != null) {
					length = (int) FcpCommand.parseHex(fcpMessage.getParameter("Length"));
				} else {
					int blockCount = (int) FcpCommand.parseHex(fcpMessage.getParameter("BlockCount"));
					int blockSize = (int) FcpCommand.parseHex(fcpMessage.getParameter("BlockSize"));
					length = blockCount * blockSize;
				}
				byte[] buffer = new byte[length];
				int bufferPos = 0;
				int read = 0;
				while (bufferPos < length) {
					read = inputStream.read(buffer, bufferPos, length - bufferPos);
					bufferPos += read;
				}
				fcpMessage.setData(buffer);
				finished = true;
			} else {
				if (fcpMessage.getName() == null) {
					fcpMessage.setName(line);
				} else {
					Parameter parameter = split(line);
					fcpMessage.addParameter(parameter.getKey(), parameter.getValue());
				}
			}
		} while (!finished);
		return fcpMessage;
	}

	private static String readLine(InputStream inputStream) throws IOException {
		StringBuffer resultBuffer = new StringBuffer();
		int b = 0;
		while (((b = inputStream.read()) != -1) && (b != 10)) {
			resultBuffer.append((char) b);
		}
		if ((resultBuffer.length() == 0) && (b == -1)) {
			return null;
		}
		return resultBuffer.toString();
	}

	private static Parameter split(String line) {
		int p = line.indexOf('=');
		String key = line.substring(0, p);
		String value = line.substring(p + 1);
		return new Parameter(key, value);
	}

	private static class Parameter {

		private String key;
		private String value;

		public Parameter(String key, String value) {
			this.key = key;
			this.value = value;
		}

		/**
		 * @return Returns the key.
		 */
		public String getKey() {
			return key;
		}

		/**
		 * @return Returns the value.
		 */
		public String getValue() {
			return value;
		}

	}

}
