/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package de.todesbaum.util.splash;

import java.awt.AWTException;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.awt.image.ImageObserver;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.lang.reflect.Method;

/**
 * Shows a splash screen while an application is loading. Usage is pretty simple:
 * 
 * <pre>
 * Splash splash = new Splash(&quot;/images/splash.png&quot;);
 * initializeApplication();
 * splash.setVisible(false, delay);
 * </pre>
 * 
 * @author David Roden &lt;droden@gmail.com&gt;
 * @version $Id:Splash.java 210 2006-03-06 13:58:28Z  $
 */
public class Splash extends Window {

	/** The image to show. */
	private Image splashImage;
	
	/** The background on screen to paint on. */
	private Image bufferImage;

	/** The width of the image. */
	private int imageWidth;

	/** The height of the image. */
	private int imageHeight;

	/**
	 * Creates a new splash screen with the image specified by <code>imageResourceName</code>. The image is loaded by means of
	 * {@link Class#getResourceAsStream(java.lang.String)} so make sure it can be found accordingly. Alsoo, it has to be in a format
	 * {@link Toolkit#createImage(byte[])} understands.
	 * 
	 * @param imageResourceName
	 *            The resource name of the image to load
	 * @see Class#getResourceAsStream(java.lang.String)
	 * @see Toolkit#createImage(byte[])
	 */
	public Splash(String imageResourceName) {
		super(new Frame());
		if (loadImage(imageResourceName)) {
			showImage();
		}
	}

	/**
	 * Tries to load the image specified by <code>imageResourceName</code>.
	 * 
	 * @param imageResourceName
	 *            The resource name of the image
	 * @return <code>true</code> if the image could be loaded, <code>false</code> otherwise
	 * @see Class#getResourceAsStream(java.lang.String)
	 * @see Toolkit#createImage(byte[])
	 */
	private boolean loadImage(String imageResourceName) {
		try {
			InputStream resourceStream = getClass().getResourceAsStream(imageResourceName);
			ByteArrayOutputStream byteOutput = new ByteArrayOutputStream(131072);
			int r = 0;
			while (r != -1) {
				byte[] buffer = new byte[131072];
				r = resourceStream.read(buffer);
				if (r > -1) {
					byteOutput.write(buffer);
				}
			}
			byteOutput.close();
			resourceStream.close();
			byte[] imageBytes = byteOutput.toByteArray();
			splashImage = Toolkit.getDefaultToolkit().createImage(imageBytes);
			return true;
		} catch (Throwable t) {
		}
		return false;
	}

	/**
	 * Causes the image to be loaded by asking for width and height properties.
	 */
	private void showImage() {
		splashImage.getWidth(this);
		splashImage.getHeight(this);
	}

	/**
	 * If the splash screen is not yet visible, move window to the center of the screen and set the window size accordingly. If this code is executed
	 * on a 1.5 VM, also set the always-on-top property of the window.
	 */
	private void setCorrectSize() {
		if (!isVisible()) {
			Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
			int splashX = (screen.width - imageWidth) / 2;
			int splashY = (screen.height - imageHeight) / 2;
			try {
				bufferImage = new Robot().createScreenCapture(new Rectangle(splashX, splashY, imageWidth, imageHeight));
			} catch (AWTException e) {
				bufferImage = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB);
			}
			setSize(imageWidth, imageHeight);
			setLocation(splashX, splashY);
			Graphics bufferGraphics = bufferImage.getGraphics();
			bufferGraphics.drawImage(splashImage, 0, 0, this);
			bufferGraphics.dispose();
			setBackground(new Color(0.75f, 0.75f, 0.75f, 0.0f));
			setSize(imageWidth, imageHeight);
			setLocation((screen.width - imageWidth) / 2, (screen.height - imageHeight) / 2);

			/* if we are running on a 1.5 vm, set the window to be always on top. */
			try {
				Method setAlwaysOnTopMethod = getClass().getMethod("setAlwaysOnTop", new Class[] { Boolean.TYPE });
				if (setAlwaysOnTopMethod != null) {
					setAlwaysOnTopMethod.invoke(this, new Object[] { Boolean.TRUE });
				}
			} catch (Exception e1) {
			}

			setVisible(true);
			toFront();
		}
	}

	/**
	 * Overridden to paint the splash image.
	 * 
	 * @param graphics
	 *            The {@link Graphics} to paint on
	 */
	public void paint(Graphics graphics) {
		if ((imageWidth != 0) && (imageHeight != 0)) {
			graphics.drawImage(bufferImage, 0, 0, this);
		}
	}

	/**
	 * Overridden to paint the splash image without redrawing the background.
	 * 
	 * @param graphics
	 *            The {@link Graphics} to paint on
	 */
	public void update(Graphics graphics) {
		paint(graphics);
	}

	/**
	 * This method is called by the image loader when more information about the image is available.
	 * 
	 * @param img
	 *            The image that is loaded
	 * @param infoflags
	 *            A set of flags stating which information is now available
	 * @param x
	 *            An x coordinate
	 * @param y
	 *            A y coordinate
	 * @param width
	 *            A width value
	 * @param height
	 *            A height value
	 */
	public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
		if ((infoflags & WIDTH) != 0) {
			this.imageWidth = width;
		}
		if ((infoflags & HEIGHT) != 0) {
			this.imageHeight = height;
		}
		if ((imageWidth != 0) && (imageHeight != 0) && ((infoflags & ImageObserver.ALLBITS) != 0)) {
			setCorrectSize();
		}
		return super.imageUpdate(img, infoflags, x, y, width, height);
	}

	/**
	 * Sets the visibility status of the splash screen.
	 * 
	 * @param visible
	 *            Show the window if <code>true</code>, hide it otherwise
	 * @see #setVisible(boolean, long)
	 */
	public void setVisible(boolean visible) {
		setVisible(visible, 0);
	}

	/**
	 * Sets the visibility status of the splash screen. The window is immediately shown but if <code>delay</code> is <code>&gt; 0</code> the
	 * hiding of the splash screen is delayed by <code>delay</code> milliseconds.
	 * 
	 * @param visible
	 *            The new visibility status
	 * @param delay
	 *            The delay for hiding the splash screen
	 */
	public void setVisible(boolean visible, final long delay) {
		if (!visible) {
			if (delay > 0) {
				new Thread(new Runnable() {
	
					public void run() {
						try {
							Thread.sleep(delay);
						} catch (InterruptedException e) {
						}
						setVisible(false);
						dispose();
					}
				}).start();
			} else {
				super.setVisible(false);
			}
		} else {
			super.setVisible(visible);
		}
	}

}
